import logging
import itertools
from typing import Sequence, Tuple, Union

import highdicom as hd
import numpy as np
from pydicom.dataset import Dataset

from dicomslide._channel import _get_channel_info
from dicomslide.utils import is_tiled_image


logger = logging.getLogger(__name__)


def disassemble_total_pixel_matrix(
    total_pixel_matrix: np.ndarray,
    tile_positions: Union[Sequence[Tuple[int, int]], np.ndarray],
    rows: int,
    columns: int,
) -> np.ndarray:
    """Disassemble a total pixel matrix into individual tiles.

    Parameters
    ----------
    total_pixel_matrix: numpy.ndarray
        Total pixel matrix
    tile_positions: Union[Sequence[Tuple[int, int]], numpy.ndarray]
        Zero-based (row, column) position of each tile in the total pixel matrix
    rows: int
        Number of rows per tile
    columns: int
        Number of columns per tile

    Returns
    -------
    numpy.ndarray
        Stacked image tiles

    """
    logger.debug('disassemble total pixel matrix')
    tile_shape: Tuple[int, ...]
    if total_pixel_matrix.ndim == 3:
        tile_shape = (rows, columns, total_pixel_matrix.shape[-1])
    elif total_pixel_matrix.ndim == 2:
        tile_shape = (rows, columns)
    else:
        raise ValueError(
            "Total pixel matrix has unexpected number of dimensions."
        )
    tiles = np.zeros(
        (len(tile_positions), *tile_shape),
        dtype=total_pixel_matrix.dtype
    )
    for i, (row_offset, col_offset) in enumerate(tile_positions):
        pixel_array = total_pixel_matrix[
            row_offset:(row_offset + rows),
            col_offset:(col_offset + columns),
            ...,
        ]
        tiles[
            i,
            0:pixel_array.shape[0],
            0:pixel_array.shape[1],
            ...
        ] = pixel_array.copy()
        del pixel_array
    return tiles


def assemble_total_pixel_matrix(
    tiles: Sequence[np.ndarray],
    tile_positions: Union[Sequence[Tuple[int, int]], np.ndarray],
    total_pixel_matrix_rows: int,
    total_pixel_matrix_columns: int,
) -> np.ndarray:
    """Assemble a total pixel matrix from individual tiles.

    Parameters
    ----------
    tiles: Sequence[numpy.ndarray]
        Individual image tiles
    tile_positions: Union[Sequence[Tuple[int, int]], numpy.ndarray]
        Zero-based (row, column) position of each tile in the total pixel matrix
    total_pixel_matrix_rows: int
        Number of total rows
    total_pixel_matrix_columns: int
        Number of total columns

    Returns
    -------
    numpy.ndarray
        Total pixel matrix

    """
    logger.debug('assemble total pixel matrix')
    if tiles[0].ndim == 3:
        rows, columns = tiles[0].shape[-3:-1]
        total_pixel_matrix = np.zeros(
            (
                total_pixel_matrix_rows,
                total_pixel_matrix_columns,
                tiles[0].shape[2],
            ),
            dtype=tiles[0].dtype,
        ) + tiles[0].dtype.type(255)
    else:
        rows, columns = tiles[0].shape[-2:]
        total_pixel_matrix = np.zeros(
            (
                total_pixel_matrix_rows,
                total_pixel_matrix_columns,
            ),
            dtype=tiles[0].dtype,
        )
    for i, tile in enumerate(tiles):
        row_start = tile_positions[i][0]
        col_start = tile_positions[i][1]
        row_stop = row_start + rows
        row_diff = total_pixel_matrix_rows - row_stop
        if row_diff < 0:
            frame_row_stop = row_diff
        else:
            frame_row_stop = None
        col_stop = col_start + columns
        col_diff = total_pixel_matrix_columns - col_stop
        if col_diff < 0:
            frame_col_stop = col_diff
        else:
            frame_col_stop = None
        total_pixel_matrix[
            row_start:row_stop,
            col_start:col_stop,
            ...,
        ] = tile[:frame_row_stop, :frame_col_stop, ...]

    return total_pixel_matrix


def compute_frame_positions(
    image: Dataset
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Compute the positions of frames.

    Parameters
    ----------
    image : pydicom.Dataset
        Metadata of a tiled image

    Returns
    -------
    total_pixel_matrix_positions: numpy.ndarray
        Zero-based (row, column) offset of the center of the top lefthand corner
        pixel of each frame from the origin of the total pixel matrix in
        pixel unit.  Values are unsigned integers in the range
        [0, Total Pixel Matrix Rows) and [0, Total Pixel Matrix Columns).
        The position of the top lefthand corner tile is (0, 0).
    slide_positions: numpy.ndarray
        Zero-based (x, y, z) offset of the center of the top lefthand corner
        pixel of each frame from the origin of the slide coordinate system
        (frame of reference) in millimeter unit. Values are floating-point
        numbers in the range [-inf, inf].
    channel_indices: numpy.ndarray
        Zero-based index for each frame into channels along the direction
        defined by successive items of the appropriate attribute. In case of
        a VL Whole Slide Microscopy Image, the attribute is the Optical Path
        Sequence, and in case of Segmentation, the attribute is the Segment
        Sequence.
    focal_plane_indices: numpy.ndarray
        Zero-based index for each frame into focal planes along depth direction
        from the glass slide towards the coverslip in the slide coordinate
        system specified by the Z Offset in Slide Coordinate System attribute.
        Values are integers in the range
        [0, Total Pixel Matrix Focal Planes).

    """
    if not is_tiled_image(image):
        raise ValueError('Argument "image" is not a a tiled image.')

    sop_instance_uid = image.SOPInstanceUID
    logger.debug(
        f'compute frame positions for image "{sop_instance_uid}"'
    )
    channels, get_channel_identifier = _get_channel_info(image)
    channel_identifier_lut = {
        str(ch.channel_identifier): i
        for i, ch in enumerate(channels)
    }
    num_channels = len(channels)

    num_frames = int(getattr(image, 'NumberOfFrames', '1'))
    if hasattr(image, 'PerFrameFunctionalGroupsSequence'):
        positions = np.stack([
            np.array(
                [
                    float(
                        channel_identifier_lut[
                            get_channel_identifier(frame_item)
                        ]
                    ),
                    float(pos_item.RowPositionInTotalImagePixelMatrix) - 1.0,
                    float(pos_item.ColumnPositionInTotalImagePixelMatrix) - 1.0,
                    float(pos_item.XOffsetInSlideCoordinateSystem),
                    float(pos_item.YOffsetInSlideCoordinateSystem),
                    float(pos_item.ZOffsetInSlideCoordinateSystem),
                ]
            )
            for frame_item in image.PerFrameFunctionalGroupsSequence
            for pos_item in frame_item.PlanePositionSlideSequence
        ])
    else:
        image_origin = image.TotalPixelMatrixOriginSequence[0]
        image_orientation = (
            float(image.ImageOrientationSlide[0]),
            float(image.ImageOrientationSlide[1]),
            float(image.ImageOrientationSlide[2]),
            float(image.ImageOrientationSlide[3]),
            float(image.ImageOrientationSlide[4]),
            float(image.ImageOrientationSlide[5]),
        )
        tiles_per_column = int(
            np.ceil(image.TotalPixelMatrixRows / image.Rows)
        )
        tiles_per_row = int(
            np.ceil(image.TotalPixelMatrixColumns / image.Columns)
        )
        num_focal_planes = getattr(
            image,
            'TotalPixelMatrixFocalPlanes',
            1
        )
        num_optical_paths = getattr(
            image,
            'NumberOfOpticalPaths',
            len(image.OpticalPathSequence)
        )

        shared_fg = image.SharedFunctionalGroupsSequence[0]
        pixel_measures = shared_fg.PixelMeasuresSequence[0]
        pixel_spacing = (
            float(pixel_measures.PixelSpacing[0]),
            float(pixel_measures.PixelSpacing[1]),
        )
        spacing_between_slices = float(
            getattr(
                pixel_measures,
                'SpacingBetweenSlices',
                1.0
            )
        )
        x_offset = float(image_origin.XOffsetInSlideCoordinateSystem)
        y_offset = float(image_origin.YOffsetInSlideCoordinateSystem)

        transformer_lut = {}
        for s in range(1, num_focal_planes + 1):
            # These checks are needed for mypy to determine the correct type
            z_offset = float(s - 1) * spacing_between_slices
            transformer_lut[s] = hd.spatial.PixelToReferenceTransformer(
                image_position=(x_offset, y_offset, z_offset),
                image_orientation=image_orientation,
                pixel_spacing=pixel_spacing
            )

        rows = int(image.Rows)
        columns = int(image.Columns)
        channel_indices = np.repeat(
            np.arange(num_channels),
            repeats=int(num_frames / num_channels)
        )

        # This is ugly, but the list comprehensive speeds up the computation
        # when compared to a for loop.
        positions = np.stack([
            np.concatenate([
                np.array([channel_index - 1], dtype=float),
                np.array(
                    [(r - 1) * rows, (c - 1) * columns],
                    dtype=float
                ),
                transformer_lut[slice_index](
                    np.array(
                        [[(c - 1) * columns, (r - 1) * rows]],
                        dtype=int
                    )
                )[0]
            ])
            for channel_index, slice_index, r, c in itertools.product(
                range(1, num_optical_paths + 1),
                range(1, num_focal_planes + 1),
                range(1, tiles_per_column + 1),
                range(1, tiles_per_row + 1),
            )
        ])

    channel_indices = positions[:, 0].astype(int)
    matrix_positions = positions[:, 1:3].astype(int)
    slide_positions = positions[:, 3:].astype(float)

    focal_plane_indices = np.zeros((num_frames, ), dtype=int)
    z_offset_values = slide_positions[:, 2]
    unique_z_offset_values = np.unique(z_offset_values)
    for index, value in enumerate(unique_z_offset_values):
        focal_plane_indices[z_offset_values == value] = index

    if hasattr(image, 'TotalPixelMatrixFocalPlanes'):
        num_focal_planes = image.TotalPixelMatrixFocalPlanes
        if len(unique_z_offset_values) != num_focal_planes:
            raise ValueError(
                'Could not compute tile positions, because '
                'an unexpected number of focal planes was found.'
            )

    return (
        matrix_positions,
        slide_positions,
        channel_indices,
        focal_plane_indices,
    )



from typing import Callable, List, NamedTuple, Tuple

from pydicom.dataset import Dataset
from pydicom.uid import (
    ParametricMapStorage,
    SegmentationStorage,
    VLWholeSlideMicroscopyImageStorage,
)

from dicomslide.enum import ChannelTypes


class _ChannelInfo(NamedTuple):

    channel_type: ChannelTypes
    channel_identifier: str
    sop_class_uid: str
    sop_instance_uid: str


def _get_channel_info(image: Dataset) -> Tuple[
    List[_ChannelInfo],
    Callable[[Dataset], str]
]:
    """Get information about channels in an image.

    The exact definition of a channel depends on the DICOM Information Object
    Definition (IOD). In case of DICOM VL Whole Slide Microscopy Image, a
    channel is an optical path and individual channels are described by items
    of the Optical Path Sequence attribute. In case of DICOM Segmentation, a
    channel is a segment and individual channels are described by items of the
    Segment Sequence attribute.

    Parameters
    ----------
    image: pydicom.Dataset
        Metadata of a DICOM image

    Returns
    -------
    channel_info: List[dicomslide._channel._ChannelInfo]
        Information for each identified channel
    get_referenced_identifier: Callable[[pydicom.Dataset], str]
        Callable to determine the identifier for an individual frame, given
        an item of either the Shared Functional Groups Sequence of the
        Per-Frame Functional Groups Sequence.

    """
    sop_class_uid = image.SOPClassUID
    sop_instance_uid = image.SOPInstanceUID
    channel_identifiers: Tuple[str, ...]
    get_referenced_identifier: Callable[[Dataset], str]
    channel_type: ChannelTypes
    if sop_class_uid == VLWholeSlideMicroscopyImageStorage:
        def get_referenced_identifier(item: Dataset) -> str:
            # Optical Path Identification Macro
            return str(
                item
                .OpticalPathIdentificationSequence[0]
                .OpticalPathIdentifier
            )

        channel_type = ChannelTypes.OPTICAL_PATH
        channel_identifiers = tuple([
            str(item.OpticalPathIdentifier)
            for item in image.OpticalPathSequence
        ])

    elif sop_class_uid == SegmentationStorage:
        def get_referenced_identifier(item: Dataset) -> str:
            # Segmentation Macro
            return str(
                item
                .SegmentIdentificationSequence[0]
                .ReferencedSegmentNumber
            )

        channel_type = ChannelTypes.SEGMENT
        channel_identifiers = tuple([
            str(item.SegmentNumber)
            for item in image.SegmentSequence
        ])

    elif sop_class_uid == ParametricMapStorage:
        def get_referenced_identifier(item: Dataset) -> str:
            return '|'.join([
                str(item.LUTLabel)
                for item in item.RealWorldValueMappingSequence
            ])

        channel_type = ChannelTypes.PARAMETER
        shared_item = image.SharedFunctionalGroupsSequence[0]
        if hasattr(shared_item, 'RealWorldValueMappingSequence'):
            channel_identifiers = (get_referenced_identifier(shared_item), )
        elif hasattr(image, 'PerFrameFunctionalGroupsSequence'):
            channel_identifiers = tuple(set([
                get_referenced_identifier(item)
                for item in image.PerFrameFunctionalGroupsSequence
            ]))
        else:
            raise ValueError(
                'DICOM Parametric Map must include attribute '
                'Real World Value Mapping Sequence either in the'
                'Functional Groups Sequence or in the Per-Frame Functional '
                'Groups Sequence.'
            )
    else:
        raise ValueError(
            'Unsupported DICOM SOP Class "{image.SOPClassUID}".'
        )
    return (
        [
            _ChannelInfo(
                channel_type,
                identifier,
                sop_class_uid,
                sop_instance_uid
            )
            for identifier in channel_identifiers
        ],
        get_referenced_identifier
    )



import itertools
import logging
from hashlib import sha256
from typing import List, Optional, Tuple

import highdicom as hd
import numpy as np
from dicomweb_client import DICOMClient
from pydicom.dataset import Dataset
from pydicom.uid import (
    ParametricMapStorage,
    SegmentationStorage,
    VLWholeSlideMicroscopyImageStorage,
)
from scipy.ndimage import rotate

from dicomslide._channel import _get_channel_info
from dicomslide.enum import ChannelTypes
from dicomslide.matrix import TotalPixelMatrix
from dicomslide.tile import compute_frame_positions
from dicomslide.utils import _encode_dataset

logger = logging.getLogger(__name__)


class TiledImage:

    """A tiled DICOM image.

    An instance of the class represents a tiled DICOM image instance and
    provides methods for convenient and efficient access of image metadata and
    pixel data from a DICOMweb server (or another source for which the
    :class:`dicomweb_client.DICOMClient` protocol has been implemented).

    A tiled image is hereby defined as a DICOM image instance that contains
    the Total Pixel Matrix Rows and Total Pixel Matrix Columns attributes.

    The class is designed to be independent of a particular DICOM Information
    Object Definition (IOD) or SOP Class and support various different types of
    DICOM images, including VL Whole Slide Microscopy Image, Segmentation, and
    Parametric Map.

    Each image is associated with one or more
    :class:`dicomslide.TotalPixelMatrix` instances, one for each unique
    combination of channel and focal plane.
    The definition of a channel is specific to a particular IOD. For example,
    in case of VL Whole Slide Microscopy Image, a channel corresponds to an
    optical path, whereas in case of a Segmentation, a channel corresponds to a
    segment.

    Examples
    --------
    >>> image = TiledImage(...)
    >>> print(image.metadata)  # pydicom.Dataset
    >>> print(image.metadata.BitsAllocated)
    >>> print(image.metadata.TotalPixelMatrixRows)
    >>> pixel_matrix = image.get_tota_pixel_matrix(channel_index=0)
    >>> print(pixel_matrix.dtype)
    >>> print(pixel_matrix.shape)
    >>> print(pixel_matrix[:1000, 350:750, :])  # numpy.ndarray

    """

    def __init__(
        self,
        client: DICOMClient,
        image_metadata: Dataset,
        max_frame_cache_size: int = 6
    ) -> None:
        """Construct object.

        Parameters
        ----------
        client: dicomweb_client.api.DICOMClient
            DICOMweb client
        image_metadata: pydicom.dataset.Dataset
            Metadata of a tiled DICOM image
        max_frame_cache_size: int, optional
            Maximum number of frames that should be cached to avoid repeated
            retrieval requests

        Note
        ----
        If `image_metadata` is the metadata of a color image, it should contain
        the ICC Profile element to enable color management. The value of this
        element may be considered bulkdata and therefore may have to be
        retrieved separately over DICOMweb.

        """
        if not isinstance(client, DICOMClient):
            raise TypeError(
                'Argument "client" must have type dicomweb_client.DICOMClient.'
            )
        self._client = client
        if not isinstance(image_metadata, Dataset):
            raise TypeError(
                'Argument "image_metadata" must have type pydicom.Dataset.'
            )
        self._metadata = image_metadata
        # The Total Pixel Matrix Focal Planes attribute must be present in case
        # of Dimension Organization TILED_FULL and may be present otherwise.
        # If it is missing, we need to determine the position of focal planes by
        # either looking up the values in the items in the Per-Frame Functional
        # Groups Sequence attribute or by computing the values if the attribute
        # is not included in the dataset.
        try:
            self._number_of_focal_planes = getattr(
                self._metadata,
                'TotalPixelMatrixFocalPlanes'
            )
            self._frame_positions = None
        except AttributeError:
            self._frame_positions = compute_frame_positions(self._metadata)
            focal_plane_indices = self._frame_positions[3]
            self._number_of_focal_planes = len(np.unique(focal_plane_indices))

        channels, _ = _get_channel_info(self._metadata)
        self._channel_identifiers = tuple([
            ch.channel_identifier for ch in channels
        ])
        self._number_of_channels = len(self._channel_identifiers)

        self._pixel_spacing = (
            float(
                self
                .metadata
                .SharedFunctionalGroupsSequence[0]
                .PixelMeasuresSequence[0]
                .PixelSpacing[0]
            ),
            float(
                self
                .metadata
                .SharedFunctionalGroupsSequence[0]
                .PixelMeasuresSequence[0]
                .PixelSpacing[1]
            ),
        )

        iterator = itertools.product(
            range(self._number_of_channels),
            range(self._number_of_focal_planes),
        )
        self._total_pixel_matrix_lut = {
            (channel_index, focal_plane_index): TotalPixelMatrix(
                client=self._client,
                image_metadata=self._metadata,
                channel_index=channel_index,
                focal_plane_index=focal_plane_index,
                max_frame_cache_size=max_frame_cache_size
            )
            for channel_index, focal_plane_index in iterator
        }

        encoded_dataset = _encode_dataset(self._metadata)
        self._quickhash = sha256(encoded_dataset).hexdigest()

        image_orientation = self.metadata.ImageOrientationSlide
        image_origin = self.metadata.TotalPixelMatrixOriginSequence[0]
        image_position = (
            float(image_origin.XOffsetInSlideCoordinateSystem),
            float(image_origin.YOffsetInSlideCoordinateSystem),
            0,  # assume that focal planes are parallel to slide surface
        )
        self._ref2pix_transformer = hd.spatial.ReferenceToPixelTransformer(
            image_orientation=image_orientation,
            image_position=image_position,
            pixel_spacing=self._pixel_spacing,
        )
        self._pix2ref_transformer = hd.spatial.PixelToReferenceTransformer(
            image_orientation=image_orientation,
            image_position=image_position,
            pixel_spacing=self._pixel_spacing,
        )

    def __hash__(self) -> int:
        return hash(self._quickhash)

    @property
    def metadata(self) -> Dataset:
        """pydicom.dataset.Dataset: Image metadata"""
        return self._metadata

    @property
    def num_channels(self) -> int:
        """int: Number of channels"""
        return self._number_of_channels

    def map_pixel_indices_to_slide_coordinates(
        self,
        pixel_indices: np.ndarray,
    ) -> np.ndarray:
        """Map pixel indices to slide coordinates.

        Parameters
        ----------
        pixel_indices: numpy.ndarray
            Zero-based (row, column) indices into the total pixel matrix of the
            image

        Returns
        -------
        numpy.ndarray
            Zero-based (x, y, z) coordinates in the slide coordinate system in
            millimeter

        """
        coordinates = np.stack(
            [
                pixel_indices[:, 1],
                pixel_indices[:, 0],
            ],
            axis=1
        )
        return self._pix2ref_transformer(coordinates)

    def get_slide_offset(
        self,
        pixel_indices: Tuple[int, int],
    ) -> Tuple[float, float]:
        """Get slide coordinates for a given total pixel matrix position.

        Parameters
        ----------
        pixel_indices: Tuple[int, int]
            Zero-based (row, column) offset in the total pixel matrix

        Returns
        -------
        Tuple[float, float]
            Zero-based (x, y) position on the slide in the slide coordinate
            system in millimeter

        """
        row_index, column_index = pixel_indices
        slide_coordinates = self._pix2ref_transformer(
            np.array([[column_index, row_index]])
        )
        return (slide_coordinates[0][0], slide_coordinates[0][1])

    def map_slide_coordinates_to_pixel_indices(
        self,
        slide_coordinates: np.ndarray,
    ) -> np.ndarray:
        """Map slide coordinates to pixel indices.

        Parameters
        ----------
        slide_coordinates: numpy.ndarray
            Zero-based (x, y, z) coordinates in the slide coordinate system in
            millimeter

        Returns
        -------
        numpy.ndarray
            Zero-based (row, column) indices into the total pixel matrix of the
            image

        """
        pixel_coordinates = self._ref2pix_transformer(slide_coordinates)
        return np.stack(
            [
                pixel_coordinates[:, 1],
                pixel_coordinates[:, 0],
            ],
            axis=1
        )

    def get_pixel_indices(
        self,
        offset: Tuple[float, float],
    ) -> Tuple[int, int]:
        """Get indices into total pixel matrix for a given slide position.

        Parameters
        ----------
        offset: Tuple[float, float]
            Zero-based (x, y) offset in the slide coordinate system in
            millimeter

        Returns
        -------
        Tuple[int, int]
            Zero-based (row, column) position in the total pixel matrix

        Note
        ----
        Pixel position may be negativ or extend beyond the size of the total
        pixel matrix if slide position at `offset` does fall into a region on
        the slide that was not imaged.

        """
        slide_coordinates = np.array([[offset[0], offset[1], 0.0]])
        pixel_indices = self._ref2pix_transformer(slide_coordinates)
        return (
            int(pixel_indices[0, 1]),
            int(pixel_indices[0, 0]),
        )

    def get_rotation(self) -> float:
        """Get angle to rotate image such that it aligns with slide.

        We want to align the image with the slide coordinate system such that
        the axes of the total pixel matrix are aligned with the X and Y axes
        of the slide coordinate system to ensure that spatial coordinates of
        graphic region of interest (ROI) annotations and are aligned with the
        source image region.

        Returns
        -------
        float
            Counterclockwise angle of rotation

        """
        image_orientation = self.metadata.ImageOrientationSlide
        radians = np.arctan2(-image_orientation[3], image_orientation[0])
        degrees = radians * 180.0 / np.pi
        degrees -= 180.0

        # Images are expected to be rotated in plane parallel to the slide
        # surface and the rows and columns of the image are expected to be
        # parallel to the axes of the slide.
        if degrees not in (-0.0, -90.0, -180.0, -270.0):
            logger.warning(
                'encountered unexpected image orientation: '
                f'{image_orientation}: {degrees}'
            )

        return degrees

    @property
    def frame_of_reference_uid(self) -> str:
        """str: Unique identifier of the frame of reference"""
        return str(self._metadata.FrameOfReferenceUID)

    @property
    def physical_offset(self) -> Tuple[float, float]:
        """Tuple[float, float]: Offset of the total pixel matrix from the
        origin of the frame of reference along the X and Y axes of the slide
        coordinate system in millimeter

        """
        image_origin = self._metadata.TotalPixelMatrixOriginSequence[0]
        return (
            float(image_origin.XOffsetInSlideCoordinateSystem),
            float(image_origin.YOffsetInSlideCoordinateSystem),
        )

    @property
    def physical_size(self) -> Tuple[float, float]:
        """Tuple[float, float]: Size of the total pixel matrix along the X and
        Y axes of the slide coordinate system in millimeter

        """
        x_endpoint, y_endpoint = self.get_slide_offset(pixel_indices=self.size)
        x_offset, y_offset = self.physical_offset
        return (
            abs(x_endpoint - x_offset),
            abs(y_endpoint - y_offset),
        )

    @property
    def size(self) -> Tuple[int, int]:
        """Tuple[int, int]: Number of total pixel matrix rows and columns"""
        return (
            int(self._metadata.TotalPixelMatrixRows),
            int(self._metadata.TotalPixelMatrixColumns),
        )

    def get_references(
        self,
        sop_class_uid: Optional[str] = None
    ) -> List[Tuple[str, str, str]]:
        """Get unique identifiers of referenced instances.

        Parameters
        ----------
        sop_class_uid: str
            SOP Class UID of instances for which references should be obtained

        Returns
        -------
        List[Tuple[str, str, str]]
            Study, Series, and SOP Instance UID of each referenced image

        """
        uids: List[Tuple[str, str, str]] = []
        study_instance_uid = self._metadata.StudyInstanceUID
        for series in getattr(self._metadata, 'ReferencedSeriesSequence', []):
            series_instance_uid = series.SeriesInstanceUID
            for instance in series.ReferencedInstanceSequence:
                sop_instance_uid = instance.ReferencedSOPInstanceUID
                if sop_class_uid is not None:
                    if sop_class_uid != instance.ReferencedSOPClassUID:
                        continue
                uids.append(
                    (
                        study_instance_uid,
                        series_instance_uid,
                        sop_instance_uid,
                    )
                )
        return uids

    def get_channel_index(self, channel_identifier: str) -> int:
        """Get index of a channel.

        The nature of the channel is specific to the SOP Class for the image.
        For example, in case of DICOM VL Whole Slide Microscopy Image, a
        channel is an optical path and in case of a DICOM Segmentation, a
        channel is a segment.

        Parameters
        ----------
        channel_identifier: str
            Identifier of a channel

        Returns
        -------
        int
            Zero-based index into channels along the direction defined by
            successive items of the corresponding attribute.

        Raises
        ------
        ValueError
            When no channel is found for `channel_identifier`

        """
        try:
            index = self._channel_identifiers.index(channel_identifier)
        except ValueError:
            raise ValueError(
                f'Image "{self._metadata.SOPInstanceUID}" does not have an '
                f'optical path with identifier "{channel_identifier}".'
            )
        return index

    @property
    def channel_type(self) -> ChannelTypes:
        """dicomslide.ChannelTypes: type of channels"""
        sop_class_uid = self._metadata.SOPClassUID
        sop_instance_uid = self._metadata.SOPInstanceUID
        lut = {
            VLWholeSlideMicroscopyImageStorage: ChannelTypes.OPTICAL_PATH,
            SegmentationStorage: ChannelTypes.SEGMENT,
            ParametricMapStorage: ChannelTypes.PARAMETER,
        }
        try:
            return lut[sop_class_uid]
        except IndexError:
            raise ValueError(
                'Could not determine channel type for '
                f'image "{sop_instance_uid}" of SOP Class "{sop_class_uid}".'
            )

    def get_channel_identifier(self, channel_index: int) -> str:
        """Get identifier of a channel.

        Parameters
        ----------
        channel_index: int
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute.

        Returns
        -------
        str
            Channel identifier

        Raises
        ------
        ValueError
            When no channel is found for `channel_index`

        """
        try:
            return self._channel_identifiers[channel_index]
        except IndexError:
            raise ValueError(
                f'Image "{self._metadata.SOPInstanceUID}" does not have a '
                f'channel {channel_index}.'
            )

    @property
    def num_focal_planes(self) -> int:
        """int: Number of focal planes"""
        return self._number_of_focal_planes

    def get_focal_plane_index(self, focal_plane_offset: float) -> int:
        """Get index of a focal plane.

        Parameters
        ----------
        focal_plane_offset: float
            Offset of the focal plane from the from the slide surface along the
            Z axis of the slide coordinate system in micrometers

        Returns
        -------
        int
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute.
            Values must be in the range [1, Total Pixel Matrix Focal Planes]

        Raises
        ------
        ValueError
            When no focal plane is found for `focal_plane_offset`

        """
        # Computing the focal plane index may be expensive, so only do it if
        # there are multiple focal planes.
        if self.num_focal_planes == 1:
            return 0
        if self._frame_positions is None:
            self._frame_positions = compute_frame_positions(self._metadata)
        slide_positions = self._frame_positions[1]
        focal_plane_indices = self._frame_positions[3]
        index = slide_positions[:, 2] == focal_plane_offset
        if np.sum(index) == 0:
            raise ValueError(
                f'Image "{self._metadata.SOPInstanceUID}" does not have a '
                f'focal plane with z offset {focal_plane_offset}.'
            )
        return focal_plane_indices[index][0]

    def get_focal_plane_offset(self, focal_plane_index: int) -> float:
        """Get z offset in slide coordinate system of a focal plane.

        Parameters
        ----------
        focal_plane_index: int
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute.
            Values must be in the range [0, Total Pixel Matrix Focal Planes).

        Returns
        -------
        float
            Offset of the focal plane from the from the slide surface along the
            Z axis of the slide coordinate system in micrometers

        Raises
        ------
        ValueError
            When no focal plane is found for `focal_plane_index`

        """
        # TODO: cache focal plane offsets
        origin_item = self._metadata.TotalPixelMatrixOriginSequence[0]
        dim_org_type = hd.DimensionOrganizationTypeValues(
            getattr(
                self._metadata,
                'DimensionOrganizationType',
                hd.DimensionOrganizationTypeValues.TILED_SPARSE.value
            )
        )
        pixel_measures_item = (
            self._metadata
            .SharedFunctionalGroupsSequence[0]
            .PixelMeasuresSequence[0]
        )
        if dim_org_type == hd.DimensionOrganizationTypeValues.TILED_FULL:
            # In case of Dimension Organization Type TILED_FULL, the absolute
            # position of focal planes is not explicitly specified and the
            # plane is assumed to be located at the slide surface.
            z_offset = float(
                getattr(
                    origin_item,
                    'ZOffsetInSlideCoordinateSytem',
                    '0.0'
                )
            )
            spacing_between_slices = float(
                getattr(pixel_measures_item, 'SpacingBetweenSlices', 1.0)
            )
            num_slices = self._metadata.TotalPixelMatrixFocalPlanes
            if focal_plane_index > num_slices:
                raise ValueError(
                    f'Image "{self._metadata.SOPInstanceUID}" does not have a '
                    f'focal plane with index {focal_plane_index}.'
                )
            if num_slices == 1:
                return z_offset
            else:
                # Note that Spacing Between Slices has millimeter unit while
                # the Z Offset in Slide Coordinate System has micrometer unit.
                return z_offset + (spacing_between_slices * num_slices * 10**3)
        else:
            if self._frame_positions is None:
                self._frame_positions = compute_frame_positions(self._metadata)
            slide_positions = self._frame_positions[1]
            focal_plane_indices = self._frame_positions[3]
            index = focal_plane_indices == focal_plane_index
            if np.sum(index) == 0:
                raise ValueError(
                    f'Image "{self._metadata.SOPInstanceUID}" does not have a '
                    f'focal plane with index {focal_plane_index}.'
                )
            return slide_positions[index, 2][0]

    def get_total_pixel_matrix(
        self,
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> TotalPixelMatrix:
        """Get total pixel matrix for a given optical path and focal plane.

        Parameters
        ----------
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute.
            Values must be in the range [0, Total Pixel Matrix Focal Planes).

        Returns
        -------
        dicomslide.TotalPixelMatrix
            Total Pixel Matrix

        """
        if (
            channel_index < 0 or
            channel_index >= self.num_channels
        ):
            raise ValueError(
                'Argument "channel_index" must be a zero-based index '
                f'in range [0, {self.num_channels}).'
            )
        if (
            focal_plane_index < 0 or
            focal_plane_index >= self.num_focal_planes
        ):
            raise ValueError(
                'Argument "focal_plane_index" must be a zero-based index '
                f'in range [0, {self.num_focal_planes}).'
            )

        key = (channel_index, focal_plane_index)
        try:
            return self._total_pixel_matrix_lut[key]
        except KeyError:
            raise ValueError(
                'Could not find a total pixel matrix for '
                f'optical path {channel_index} and '
                f'focal plane {focal_plane_index}.'
            )

    def get_image_region(
        self,
        offset: Tuple[int, int],
        size: Tuple[int, int],
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> np.ndarray:
        """Get image region.

        Parameters
        ----------
        offset: Tuple[int, int]
            Zero-based (row, column) indices in the range [0, Rows) and
            [0, Columns), respectively, that specify the offset of the image
            region in the total pixel matrix. The ``(0, 0)`` coordinate is
            located at the center of the topleft hand pixel in the total pixel
            matrix.
        size: Tuple[int, int]
            Rows and columns of the requested image region
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute.
            Values must be in the range [0, Total Pixel Matrix Focal Planes)

        Returns
        -------
        numpy.ndarray
            Three-dimensional pixel array of shape
            (Rows, Columns, Samples per Pixel) for the requested image region

        """
        matrix = self.get_total_pixel_matrix(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )

        row_start, col_start = offset
        rows, cols = size
        row_stop = row_start + rows
        col_stop = col_start + cols
        logger.debug(
            f'get region [{row_start}:{row_stop}, {col_start}:{col_stop}, :] '
            f'for channel {channel_index} and focal plane {focal_plane_index} '
            f'of image "{self._metadata.SOPInstanceUID}"'
        )
        return matrix[row_start:row_stop, col_start:col_stop, :]

    def get_slide_region(
        self,
        offset: Tuple[float, float],
        size: Tuple[float, float],
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> np.ndarray:
        """Get slide region.

        Parameters
        ----------
        offset: Tuple[float, float]
            Zero-based (x, y) offset in the slide coordinate system in
            millimeter resolution. The ``(0.0, 0.0)`` coordinate is located at
            the origin of the slide (usually the slide corner).
        size: Tuple[float, float]
            Width and height of the requested slide region in millimeter unit
            along the X and Y axis of the slide coordinate system, respectively.
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute.
            Values must be in the range [0, Total Pixel Matrix Focal Planes)

        Returns
        -------
        numpy.ndarray
            Three-dimensional pixel array of shape
            (Rows, Columns, Samples per Pixel) for the requested slide region

        Note
        ----
        The slide coordinate system is defined for the upright standing slide
        such that the X axis corresponds to the short side of the slide and the
        Y axis corresponds to the long side of the slide.
        The rows of the returned pixel array are thus parallel to the X axis of
        the slide coordinate system and the columns parallel to the Y axis of
        the slide coordinate system.

        """
        logger.debug(
            f'get region on slide of size {size} [mm] at offset {offset} [mm] '
            f'for channel {channel_index} and focal plane {focal_plane_index} '
            f'of image "{self._metadata.SOPInstanceUID}"'
        )
        matrix = self.get_total_pixel_matrix(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )

        pixel_indices = np.array([
            self.get_pixel_indices(offset=offset),
            self.get_pixel_indices(
                offset=(
                    offset[0] + size[0],
                    offset[1] + size[1],
                )
            )
        ])
        row_index = np.min(pixel_indices[:, 0])
        col_index = np.min(pixel_indices[:, 1])

        region_rows = int(np.ceil(size[1] / self._pixel_spacing[1]))
        region_cols = int(np.ceil(size[0] / self._pixel_spacing[0]))

        # Region may extend beyond the image's total pixel matrix
        col_overhang = abs(min([col_index, 0]))
        col_start = min([
            max([col_index, col_overhang]),
            matrix.shape[1] - 1,
        ])
        cols = min([
            region_cols - col_overhang - 1,
            matrix.shape[1] - col_start - 1,
        ])
        col_stop = col_start + cols
        region_col_start = col_overhang
        region_col_stop = region_col_start + cols

        col_diff = (region_col_stop - region_col_start) - (col_stop - col_start)
        if col_diff != 0:
            raise ValueError(
                'Failed to get slide region: '
                'Could not determine the number of columns.'
            )

        row_overhang = abs(min([row_index, 0]))
        row_start = min([
            max([row_index, row_overhang]),
            matrix.shape[0] - 1,
        ])
        rows = min([
            region_rows - row_overhang - 1,
            matrix.shape[0] - row_start - 1,
        ])
        row_stop = row_start + rows
        region_row_start = row_overhang
        region_row_stop = region_row_start + rows

        row_diff = (region_row_stop - region_row_start) - (row_stop - row_start)
        if row_diff != 0:
            raise ValueError(
                'Failed to get slide region: '
                'Could not determine the number of rows.'
            )

        region = np.zeros(
            (region_rows, region_cols, matrix.shape[2]),
            dtype=matrix.dtype
        )
        if self.metadata.SamplesPerPixel == 3:
            region += 255

        try:
            region[
                region_row_start:region_row_stop,
                region_col_start:region_col_stop
            ] = matrix[row_start:row_stop, col_start:col_stop, :]
        except ValueError as error:
            raise ValueError(f'Failed to get slide region: {error}')

        degrees = self.get_rotation()

        return rotate(region, angle=degrees)



from io import BytesIO
from typing import Optional, Union

import highdicom as hd
from pydicom.dataset import Dataset
from pydicom.filewriter import dcmwrite
from pydicom.sr.coding import Code

from dicomslide.enum import ImageFlavors


def is_image(dataset: Dataset) -> bool:
    """Determine whether a dataset is an image.

    Parameters
    ----------
    dataset: pydicom.dataset.Dataset
        Dataset

    Returns
    -------
    bool
        Whether dataset is an image

    """
    return all([
        hasattr(dataset, 'BitsAllocated'),
        hasattr(dataset, 'Columns'),
        hasattr(dataset, 'Rows'),
        hasattr(dataset, 'PhotometricInterpretation'),
    ])


def is_tiled_image(dataset: Dataset) -> bool:
    """Determine whether a dataset is a tiled image.

    Parameters
    ----------
    dataset: pydicom.dataset.Dataset
        Dataset

    Returns
    -------
    bool
        Whether dataset is a tiled image

    """
    if is_image(dataset):
        return all([
            hasattr(dataset, 'TotalPixelMatrixColumns'),
            hasattr(dataset, 'TotalPixelMatrixRows'),
        ])
    return False


def is_volume_image(dataset: Dataset) -> bool:
    """Determine whether a dataset is a VOLUME or THUMBNAIL image.

    Parameters
    ----------
    dataset: pydicom.dataset.Dataset
        Dataset

    Returns
    -------
    bool
        Whether dataset is a VOLUME or THUMBNAIL image

    """
    if is_image(dataset):
        try:
            return dataset.ImageType[2] in (
                ImageFlavors.VOLUME.value,
                ImageFlavors.THUMBNAIL.value,
            )
        except IndexError:
            return True
    return False


def is_label_image(dataset: Dataset) -> bool:
    """Determine whether a dataset is a LABEL image.

    Parameters
    ----------
    dataset: pydicom.dataset.Dataset
        Dataset

    Returns
    -------
    bool
        Whether dataset is a LABEL image

    """
    if is_image(dataset):
        try:
            return dataset.ImageType[2] == ImageFlavors.LABEL.value
        except IndexError:
            return False
    return False


def is_overview_image(dataset: Dataset) -> bool:
    """Determine whether a dataset is an OVERVIEW image.

    Parameters
    ----------
    dataset: pydicom.dataset.Dataset
        Dataset

    Returns
    -------
    bool
        Whether dataset is an OVERVIEW image

    """
    if is_image(dataset):
        try:
            return dataset.ImageType[2] == ImageFlavors.OVERVIEW.value
        except IndexError:
            return False
    return False


def _encode_dataset(dataset: Dataset) -> bytes:
    """Encode DICOM dataset.

    Parameters
    ----------
    dataset: pydicom.dataset.Dataset
        Dataset

    Returns
    -------
    bytes
        Binary encoded dataset

    """
    clone = Dataset(dataset)
    clone.is_little_endian = True
    clone.is_implicit_VR = False
    # Only encode the Data Set.
    if hasattr(clone, 'file_meta'):
        del clone.file_meta
    # Some servers may erroursly include these Data Elements into the Data Set
    # instead of the File Meta Information.
    for element in clone.group_dataset(0x0002):
        del clone[element.tag]
    with BytesIO() as fp:
        dcmwrite(fp, clone, write_like_original=True)
        encoded_value = fp.getvalue()
    return encoded_value


def does_optical_path_item_match(
    item: Dataset,
    identifier: Optional[str] = None,
    description: Optional[str] = None,
    illumination_wavelength: Optional[float] = None,
) -> bool:
    """Check whether an optical path item matches.

    Parameters
    ----------
    item: pydicom.Dataset
        Item of the Optical Path Sequence
    identifier: Union[str, None], optional
        Optical path identifier
    description: Union[str, None], optional
        Optical path description
    illumination_wavelength: Union[float, None], optional
        Illumination wave length

    Returns
    -------
    bool
        Whether item matches

    """  # noqa: E501
    matches = []
    if identifier is not None:
        matches.append(item.OpticalPathIdentifier == identifier)
    if description is not None:
        if hasattr(item, 'OpticalPathDescription'):
            matches.append(item.OpticalPathDescription == description)
        else:
            matches.append(False)
    if illumination_wavelength is not None:
        if hasattr(item, 'IlluminationWaveLength'):
            matches.append(
                item.IlluminationWaveLength == illumination_wavelength
            )
        else:
            matches.append(False)
    if len(matches) == 0:
        return True
    return any(matches)


def does_specimen_description_item_match(
    item: Dataset,
    specimen_stain: Optional[Union[hd.sr.CodedConcept, Code]] = None
) -> bool:
    """Check whether a specimen description item matches.

    Parameters
    ----------
    item: pydicom.Dataset
        Item of the Specimen Description Sequence
    specimen_stain: Union[pydicom.sr.coding.Code, highdicom.sr.CodedConcept, None], optional
        Specimen stain substance

    Returns
    -------
    bool
        Whether item matches

    """  # noqa: E501
    matches = []
    if specimen_stain is not None:
        description = hd.SpecimenDescription.from_dataset(item)
        is_specimen_staining_described = False
        for step in description.specimen_preparation_steps:
            procedure = step.processing_procedure
            if isinstance(procedure, hd.SpecimenStaining):
                is_specimen_staining_described = True
                substances = [
                    item
                    for item in procedure.substances
                    if isinstance(item, hd.sr.CodedConcept)
                ]
                matches.append(specimen_stain in substances)
        if not is_specimen_staining_described:
            matches.append(False)
    if len(matches) == 0:
        return True
    return any(matches)


def does_segment_item_match(
    item: Dataset,
    number: Optional[int] = None,
    label: Optional[str] = None,
    property_category: Optional[Union[hd.sr.CodedConcept, Code]] = None,
    property_type: Optional[Union[hd.sr.CodedConcept, Code]] = None,
) -> bool:
    """Check whether a segment item matches.

    Parameters
    ----------
    item: pydicom.Dataset
        Item of the Segment Sequence
    number: Union[int, None], optional
        Segment number
    label: Union[int, None], optional
        Segment label
    property_category: Union[pydicom.sr.coding.Code, highdicom.sr.CodedConcept, None], optional
        Segmented property category
    property_type: Union[pydicom.sr.coding.Code, highdicom.sr.CodedConcept, None], optional
        Segmented property type

    Returns
    -------
    bool
        Whether item matches

    """  # noqa: E501
    matches = []
    if number is not None:
        matches.append(item.SegmentNumber == number)
    if label is not None:
        matches.append(item.SegmentLabel == label)
    if property_category is not None:
        code_item = hd.sr.CodedConcept.from_dataset(
            item.SegmentedPropertyCategoryCodeSequence[0]
        )
        matches.append(code_item == property_category)
    if property_type is not None:
        code_item = hd.sr.CodedConcept.from_dataset(
            item.SegmentedPropertyTypeCodeSequence[0]
        )
        matches.append(code_item == property_type)
    if len(matches) == 0:
        return True
    return any(matches)



from enum import Enum


class ImageFlavors(Enum):

    """Enumerated values for image flavors."""

    VOLUME = 'VOLUME'
    LABEL = 'LABEL'
    OVERVIEW = 'OVERVIEW'
    THUMBNAIL = 'THUMBNAIL'


class ChannelTypes(Enum):

    """Enumerated values for channel types."""

    OPTICAL_PATH = 'OPTICAL_PATH'
    SEGMENT = 'SEGMENT'
    PARAMETER = 'PARAMETER'



import logging
import sys
import traceback
from typing import Dict, Tuple

import numpy as np
from PIL import Image as PillowImage

from dicomslide.enum import ImageFlavors
from dicomslide.slide import Slide

logger = logging.getLogger(__name__)

OPENSLIDE_COMMENT = 'openslide.comment'
OPENSLIDE_VENDOR = 'openslide.vendor'
OPENSLIDE_QUICKHASH1 = 'openslide.quickhash-1'
OPENSLIDE_BACKGROUND_COLOR = 'openslide.background-color'
OPENSLIDE_OBJECTIVE_POWER = 'openslide.objective-power'
OPENSLIDE_MPP_X = 'openslide.mpp-x'
OPENSLIDE_MPP_Y = 'openslide.mpp-y'
OPENSLIDE_BOUNDS_X = 'openslide.bounds-x'
OPENSLIDE_BOUNDS_Y = 'openslide.bounds-y'
OPENSLIDE_BOUNDS_WIDTH = 'openslide.bounds-width'
OPENSLIDE_BOUNDS_HEIGHT = 'openslide.bounds-height'


class OpenSlide:

    """Wrapper class that exposes data of a slide via the OpenSlide interface.

    Note
    ----
    There are two major differences between the OpenSlide interface exposed by
    this class and the interface exposed by :class:`dicomslide.Slide`::

        1. The OpenSlide API returns images as :class:`PIL.Image.Image` objects,
        while :class:`dicomslide.Slide` returns pixel arrays as
        :class:`numpy.ndarray` objects.
        2. The OpenSlide API specifies image dimensions and indices in
        column-major order (following the Pillow convention), while
        :class:`dicomslide.Slide` specifies array dimensions and indices in
        row-major order (following the NumPy convention).

    """

    def __init__(self, slide: Slide):
        """

        Parameters
        ----------
        slide: dicomslide.slide.Slide
            DICOM slide

        """
        self._slide = slide
        self._volume_images = self._slide.get_volume_images()
        for image in self._volume_images:
            if image.metadata.SamplesPerPixel != 3:
                raise ValueError(
                    'OpenSlide API only supports color images.'
                )
            if image.num_channels > 1:
                raise ValueError(
                    'OpenSlide API only supports images with a single '
                    'optical path.'
                )
            if image.num_focal_planes > 1:
                raise ValueError(
                    'OpenSlide API only supports images with a single '
                    'focal plane.'
                )

    @property
    def level_count(self) -> int:
        """int: Number of pyramid resolution levels"""
        return self._slide.num_levels

    @property
    def dimensions(self) -> Tuple[int, int]:
        """Tuple[int, int]: Width and height of images at base level 0"""
        return (
            self._slide.total_pixel_matrix_dimensions[0][1],
            self._slide.total_pixel_matrix_dimensions[0][0],
        )

    @property
    def level_dimensions(self) -> Tuple[Tuple[int, int], ...]:
        """Tuple[Tuple[int, int]]: Width and height of images at each level"""
        return tuple([
            (dimensions[1], dimensions[0])
            for dimensions in self._slide.total_pixel_matrix_dimensions
        ])

    @property
    def level_downsamples(self) -> Tuple[float, ...]:
        """Tuple[float]: Downsampling factor of images at each level with
        respect to the base level 0
        """
        return self._slide.downsampling_factors

    def _convert_to_pil_image(
        self,
        pixel_array: np.ndarray,
    ) -> PillowImage.Image:
        """Decode a frame and convert it into a Pillow Image object.

        Parameters
        ----------
        frame: bytes
            Frame item of the Pixel Data element of an image
        transfer_syntax_uid: str
            UID of the transfer syntax

        Returns
        -------
        PIL.Image.Image
            RGBA image

        """
        shape = pixel_array.shape
        image = PillowImage.frombuffer(
            data=pixel_array,
            size=(shape[1], shape[0]),
            mode='RGB'
        )
        return image.convert('RGBA')

    @property
    def associated_images(self) -> Dict[str, PillowImage.Image]:
        """Dict[str, PIL.Image.Image]: Mapping of image flavor (LABEL or
        OVERVIEW) to image

        """
        mapping = {}
        if len(self._slide.label_images) > 0:
            if len(self._slide.label_images) > 1:
                logger.warning(
                    'slide has more than one associated LABEL image'
                )
            image = self._slide.label_images[0]
            matrix = image.get_total_pixel_matrix()
            flavor = ImageFlavors.LABEL.value
            mapping[flavor] = self._convert_to_pil_image(matrix[:, :, :])
        if len(self._slide.overview_images) > 0:
            if len(self._slide.overview_images) > 1:
                logger.warning(
                    'slide has more than one associated OVERVIEW image'
                )
            image = self._slide.overview_images[0]
            matrix = image.get_total_pixel_matrix()
            flavor = ImageFlavors.OVERVIEW.value
            mapping[flavor] = self._convert_to_pil_image(matrix[:, :, :])
        return mapping

    def read_region(
        self,
        location: Tuple[int, int],
        level: int,
        size: Tuple[int, int]
    ) -> PillowImage.Image:
        """Read region of a VOLUME (or THUMBNAIL) image at a given level.

        Parameters
        ----------
        location: Tuple[int, int]
            Zero-based (column, row) offset of the region from the topleft hand
            pixel of of the total pixel matrix of the image at the base level 0
        level: int
            Zero-based level index
        size: Tuple[int, int]
            Number of pixels columns and rows that should be read from the
            total pixel matrix at the specified level

        Returns
        -------
        PIL.Image.Image
            RGBA image

        """
        pixel_array = self._slide.get_image_region(
            offset=(location[1], location[0]),
            level=level,
            size=(size[1], size[0]),
            channel_index=0,
            focal_plane_index=0
        )
        return self._convert_to_pil_image(pixel_array)

    @property
    def properties(self) -> Dict[str, str]:
        """Metadata about the slide.

        Returns
        -------
        Dict[str, str]
            OpenSlide properties

        """
        image = self._volume_images[0]
        image_metadata = image.metadata
        pixel_spacing = (
            image_metadata
            .SharedFunctionalGroupsSequence[0]
            .PixelMeasuresSequence[0]
            .PixelSpacing
        )
        # TODO: Does ImageOrientationSlide need to be taken into account?
        x_offset = (
            image_metadata
            .TotalPixelMatrixOriginSequence[0]
            .XOffsetInSlideCoordinateSystem
        )
        y_offset = (
            image_metadata
            .TotalPixelMatrixOriginSequence[0]
            .YOffsetInSlideCoordinateSystem
        )

        return {
            OPENSLIDE_COMMENT: 'DICOM',
            OPENSLIDE_VENDOR: image_metadata.Manufacturer,
            OPENSLIDE_QUICKHASH1: str(hash(self._slide)),
            # TODO: Consider using RecommendedAbsentPixelCIELabValue if
            # available. However, that would need to be used during decoding
            # when the ICC Profile is applied.
            OPENSLIDE_BACKGROUND_COLOR: 'FFFFFF',
            OPENSLIDE_OBJECTIVE_POWER: getattr(
                image_metadata.OpticalPathSequence[0],
                'ObjectiveLensPower',
                ''
            ),
            OPENSLIDE_MPP_X: str(pixel_spacing[1] * 10**3),
            OPENSLIDE_MPP_Y: str(pixel_spacing[1] * 10**3),
            OPENSLIDE_BOUNDS_X: x_offset,
            OPENSLIDE_BOUNDS_Y: y_offset,
            OPENSLIDE_BOUNDS_WIDTH: str(image.physical_size[1]),
            OPENSLIDE_BOUNDS_HEIGHT: str(image.physical_size[0]),
        }

    def get_best_level_for_downsample(self, downsample: float) -> int:
        """Compute best level for displaying the given downsample.

        Parameters
        ----------
        downsample: float
            Desired downsample factor

        Returns
        -------
        int
            Zero-based level index

        """
        level_downsamples = np.array(self.level_downsamples)
        distances = np.abs(level_downsamples - downsample)
        return int(np.argmin(distances))

    def get_thumbnail(self, size: Tuple[int, int]) -> PillowImage.Image:
        """Create a thumbnail of the slide.

        Parameters
        ----------
        size: Tuple[int, int]
            Number of pixels columns and rows that the thumbnail should have

        Returns
        -------
        PIL.Image.Image
            RGB image

        """
        downsampling_factor = max(
            *(dim / thumb for dim, thumb in zip(self.dimensions, size))
        )
        level = self.get_best_level_for_downsample(downsampling_factor)
        tile = self.read_region((0, 0), level, self.level_dimensions[level])
        background_color = '#{}'.format(
            self.properties.get(OPENSLIDE_BACKGROUND_COLOR, 'ffffff')
        )
        thumb = PillowImage.new('RGB', tile.size, background_color)
        thumb.paste(tile, None, tile)
        thumb.thumbnail(size, PillowImage.ANTIALIAS)
        return thumb

    def __enter__(self):
        return self

    def __exit__(self, except_type, except_value, except_trace) -> None:
        self.close()
        if except_value:
            sys.stderr.write('Error while reading slide:\n{except_value}')
            for tb in traceback.format_tb(except_trace):
                sys.stderr.write(tb)
            raise

    def close(self) -> None:
        # Method is only implemented for compability with the OpenSlide API.
        pass



from dicomslide.enum import ChannelTypes, ImageFlavors
from dicomslide.image import TiledImage
from dicomslide.matrix import TotalPixelMatrix, TotalPixelMatrixSampler
from dicomslide.pyramid import (
    compute_image_center_position,
    get_image_pixel_spacing,
    get_image_size,
    select_image_at_magnification,
    select_image_at_pixel_spacing,
    sort_images_by_pixel_spacing,
    sort_images_by_size,
)
from dicomslide.openslide import OpenSlide
from dicomslide.pyramid import Pyramid, PyramidLevel
from dicomslide.slide import find_slides, Slide
from dicomslide.tile import (
    assemble_total_pixel_matrix,
    compute_frame_positions,
    disassemble_total_pixel_matrix,
)
from dicomslide.utils import (
    does_optical_path_item_match,
    does_segment_item_match,
    does_specimen_description_item_match,
    is_image,
    is_label_image,
    is_overview_image,
    is_tiled_image,
    is_volume_image,
)
from dicomslide.version import __version__


__all__ = [
    'assemble_total_pixel_matrix',
    'ChannelTypes',
    'compute_frame_positions',
    'compute_image_center_position',
    'disassemble_total_pixel_matrix',
    'does_optical_path_item_match',
    'does_segment_item_match',
    'does_specimen_description_item_match',
    'find_slides',
    'get_image_pixel_spacing',
    'get_image_size',
    'ImageFlavors',
    'is_image',
    'is_label_image',
    'is_overview_image',
    'is_tiled_image',
    'is_volume_image',
    'OpenSlide',
    'Pyramid',
    'PyramidLevel',
    'select_image_at_magnification',
    'select_image_at_pixel_spacing',
    'Slide',
    'sort_images_by_pixel_spacing',
    'sort_images_by_size',
    'TiledImage',
    'TotalPixelMatrix',
    'TotalPixelMatrixSampler',
    '__version__',
]



import logging
from operator import eq
from typing import Dict, Iterator, List, NamedTuple, Optional, Sequence, Tuple

import highdicom as hd
import numpy as np
from pydicom.dataset import Dataset

from dicomslide.utils import is_tiled_image, is_volume_image

logger = logging.getLogger(__name__)


def get_image_pixel_spacing(image: Dataset) -> Tuple[float, float]:
    """Get pixel spacing (spacing between pixels) of an image.

    Parameters
    ----------
    image: pydicom.dataset.Dataset
        Metadata of a DICOM VL Whole Slide Microscopy Image instance
        derived image instance (e.g., DICOM Segmentation)

    Returns
    -------
    Tuple[float, float]
        Pixel spacing

    Note
    ----
    It is assumed that pixels are square.

    """
    pixel_spacing = (
        image
        .SharedFunctionalGroupsSequence[0]
        .PixelMeasuresSequence[0]
        .PixelSpacing
    )
    if len(set(pixel_spacing)) > 1:
        logger.warn(
            'pixel spacing is different along row and column directions'
        )
    return (float(pixel_spacing[0]), float(pixel_spacing[1]))


def get_image_size(image: Dataset) -> int:
    """Get size of an image.

    Parameters
    ----------
    image: pydicom.dataset.Dataset
        Metadata of a DICOM VL Whole Slide Microscopy Image instance or a
        derived image instance (e.g., DICOM Segmentation)

    Returns
    -------
    int
        Number of pixels in each total pixel matrix

    """
    return int(image.TotalPixelMatrixRows) * int(image.TotalPixelMatrixColumns)


def sort_images_by_resolution(
    collection: Sequence[Dataset]
) -> List[Dataset]:
    """Sort images by resolution in descending order (highest to lowest).

    Parameters
    ----------
    collection: Sequence[pydicom.dataset.Dataset]
        Metadata of DICOM VL Whole Slide Microscopy Image instances

    Returns
    -------
    List[pydicom.dataset.Dataset]
        Sorted metadata of DICOM VL Whole Slide Microscopy Image instances

    """
    def get_image_pixel_spacing_rows(image: Dataset) -> float:
        return get_image_pixel_spacing(image)[0]

    return sorted(collection, key=get_image_pixel_spacing_rows, reverse=True)


def sort_images_by_pixel_spacing(
    collection: Sequence[Dataset]
) -> List[Dataset]:
    """Sort images by pixel spacing in ascending order (lowest to highest).

    Parameters
    ----------
    collection: Sequence[pydicom.dataset.Dataset]
        Metadata of DICOM VL Whole Slide Microscopy Image instances

    Returns
    -------
    List[pydicom.dataset.Dataset]
        Sorted metadata of DICOM VL Whole Slide Microscopy Image instances

    """
    return sorted(collection, key=get_image_pixel_spacing, reverse=False)


def sort_images_by_size(collection: Sequence[Dataset]) -> List[Dataset]:
    """Sort images by size in descending order (largest to smallest).

    Parameters
    ----------
    collection: Sequence[pydicom.dataset.Dataset]
        Metadata of DICOM VL Whole Slide Microscopy Image instances

    Returns
    -------
    List[pydicom.dataset.Dataset]
        Sorted metadata of DICOM VL Whole Slide Microscopy Image instances

    """
    return sorted(collection, key=get_image_size, reverse=True)


def select_image_at_magnification(
    collection: Sequence[Dataset],
    magnification: int,
    tolerance: Optional[float] = None,
) -> Dataset:
    """Select an image from a collection at a desired magnification.

    Parameters
    ----------
    collection: Sequence[pydicom.dataset.Dataset]
        Metadata of DICOM VL Whole Slide Microscopy Image instances
    magnification:
        Magnification level (corresponds roughly to object lens power of a
        microscope) of the image that should be selected.
        Note that an image with an exactly matching magnification may not
        exist. In this case, the nearest level will be chosen.
        Choices: ``{2, 4, 10, 20, 40}``
    tolerance: Union[float, None], optional
        Difference between target magnification and closest available
        magnification in millimeter that can be tolerated.

    Returns
    -------
    pydicom.dataset.Dataset
        Image closest to the desired magnification

    Raises
    ------
    ValueError
        When argument `collection` is an emtpy sequence, when argument
        `magnification` does not match one of the available options, or when
        `tolerance` is exceeded.

    """
    if len(collection) == 0:
        raise ValueError('Argument "collection" must not be empty.')

    magnification_to_pixel_spacing = {
        1: 0.01,
        2: 0.005,
        4: 0.0025,
        10: 0.001,
        20: 0.0005,
        40: 0.00025,
    }
    try:
        pixel_spacing = magnification_to_pixel_spacing[magnification]
    except KeyError:
        raise ValueError(
            'Argument "magnification" should be one of the following: '
            '"{}"'.format('", "'.join([
                str(key) for key in magnification_to_pixel_spacing.keys()
            ]))
        )

    return select_image_at_pixel_spacing(
        collection=collection,
        pixel_spacing=(pixel_spacing, pixel_spacing),
        tolerance=tolerance
    )


def select_image_at_pixel_spacing(
    collection: Sequence[Dataset],
    pixel_spacing: Tuple[float, float],
    tolerance: Optional[float] = None,
) -> Dataset:
    """Select an image from a collection at a desired spatial pixel spacing.

    Parameters
    ----------
    collection: Sequence[pydicom.dataset.Dataset]
        Metadata of DICOM VL Whole Slide Microscopy Image instances
    pixel_spacing: Tuple[float, float]
        Desired spacing between two pixels along the row and column direction
        of the image from top to bottom and left to right, respectively.
    tolerance: Union[float, None], optional
        Difference between target magnification and closest available
        magnification in millimeter that can be tolerated.

    Returns
    -------
    pydicom.dataset.Dataset
        Image closest to the desired pixel spacing

    Raises
    ------
    ValueError
        When argument `collection` is an emtpy sequence or when
        `tolerance` is exceeded.

    Note
    ----
    If multiple images with the same pixel spacing are contained in
    `collection`, the first matching image will be returned. It is the
    responsibility of the caller to filter the images beforehand if necessary.

    """
    if len(collection) == 0:
        raise ValueError('Argument "collection" must not be empty.')

    all_pixel_spacings = np.array([
        get_image_pixel_spacing(image)
        for image in collection
    ])
    distances = np.abs(
        np.mean(
            [
                all_pixel_spacings[:, 0] - pixel_spacing[0],
                all_pixel_spacings[:, 1] - pixel_spacing[1],
            ],
            axis=0
        )
    )
    index_nearest = int(np.argmin(distances))
    distance_nearest = distances[index_nearest]

    if tolerance is not None:
        if distance_nearest > tolerance:
            raise ValueError(
                'Could not find image with suitable pixel spacing '
                f'{pixel_spacing}. Distance between requested pixel spacing '
                'and nearest available pixel spacing exceeded '
                f'tolerance of {tolerance} mm by {distance_nearest} mm.'
            )

    return collection[index_nearest]


def compute_image_center_position(image: Dataset) -> Tuple[float, float, float]:
    """Compute position of image center in slide coordinate system.

    Parameters
    ----------
    image: pydicom.dataset.Dataset
        Metadata of DICOM VL Whole Slide Microscopy Image instance

    Returns
    -------
    Tuple[float, float, float]
        (x, y, z) coordinates

    """
    image_origin = image.TotalPixelMatrixOriginSequence[0]
    transformer = hd.spatial.ImageToReferenceTransformer(
        image_position=(
            image_origin.XOffsetInSlideCoordinateSystem,
            image_origin.YOffsetInSlideCoordinateSystem,
            0.0,
        ),
        image_orientation=image.ImageOrientationSlide,
        pixel_spacing=(
            image
            .SharedFunctionalGroupsSequence[0]
            .PixelMeasuresSequence[0]
            .PixelSpacing
        )
    )
    coordinates = transformer(
        np.array([
            (
                image.TotalPixelMatrixColumns / 2,
                image.TotalPixelMatrixRows / 2,
            )
        ])
    )
    return (coordinates[0, 0], coordinates[0, 1], coordinates[0, 2])


class PyramidLevel(NamedTuple):

    """Image pyramid level."""

    total_pixel_matrix_dimensions: Tuple[int, int]
    pixel_spacing: Tuple[float, float]
    downsampling_factors: Tuple[float, float]
    has_pixels: bool

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return all([
            eq(
                self.total_pixel_matrix_dimensions,
                other.total_pixel_matrix_dimensions
            ),
            eq(
                self.pixel_spacing,
                other.pixel_spacing
            ),
            eq(
                self.downsampling_factors,
                other.downsampling_factors
            ),
            eq(
                self.has_pixels,
                other.has_pixels
            ),
        ])

    def __contains__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return all([
            eq(
                self.total_pixel_matrix_dimensions,
                other.total_pixel_matrix_dimensions
            ),
            eq(
                self.pixel_spacing,
                other.pixel_spacing
            ),
            eq(
                self.downsampling_factors,
                other.downsampling_factors
            ),
        ])


class Pyramid:

    """Image pyramid."""

    def __init__(
        self,
        metadata: Sequence[Dataset],
        tolerance: float,
        ref_metadata: Optional[Sequence[Dataset]] = None
    ) -> None:
        """

        Parameters
        ----------
        metadata: Sequence[pydicom.Dataset]
            Metadata of DICOM image instances
        tolerance: float
            Maximally tolerated distances between the centers of images at
            different pyramid levels in the slide coordinate system in
            millimeter unit
        ref_metadata: Union[Sequence[pydicom.Dataset], None], optional
            Metadata of referenced DICOM source image instances that may serve
            as a template

        """
        if not all([is_volume_image(image) for image in metadata]):
            raise ValueError(
                'Images in pyramid must have flavor VOLUME or THUMBNAIL.'
            )

        if not all([is_tiled_image(image) for image in metadata]):
            raise ValueError('Images in pyramid must be tiled.')

        sizes = [get_image_size(image) for image in metadata]
        if len(set(sizes)) != len(sizes):
            raise ValueError('Images in pyramid must have unique sizes.')

        if not np.array_equal(
            np.argsort(sizes),
            np.flip(np.arange(len(sizes)))
        ):
            raise ValueError(
                'Images in pyramid must be sorted by size in decending order.'
            )

        coordinates = np.array([
            compute_image_center_position(image)
            for image in metadata
        ])
        max_point = np.max(coordinates, axis=0)
        min_point = np.min(coordinates, axis=0)
        distance = np.linalg.norm(max_point - min_point)
        if distance > tolerance:
            raise ValueError(
                'Images in pyramid must be spatially aligned. '
                'Distances between image centers in slide coordinate system '
                f'differ by more than {tolerance} mm: \n{coordinates}'
            )

        if ref_metadata is None:
            ref_metadata = metadata

        level_lut: Dict[Tuple[int, int], PyramidLevel] = {}
        level_keys: List[Tuple[int, int]] = []
        ref_base_image = ref_metadata[0]
        for i, ref_image in enumerate(ref_metadata):
            total_pixel_matrix_dimensions = (
                ref_image.TotalPixelMatrixRows,
                ref_image.TotalPixelMatrixColumns,
            )
            pixel_spacing = (
                float(
                    ref_image
                    .SharedFunctionalGroupsSequence[0]
                    .PixelMeasuresSequence[0]
                    .PixelSpacing[1]
                ),
                float(
                    ref_image
                    .SharedFunctionalGroupsSequence[0]
                    .PixelMeasuresSequence[0]
                    .PixelSpacing[0]
                ),
            )
            downsampling_factors = (
                (
                    ref_base_image.TotalPixelMatrixRows /
                    ref_image.TotalPixelMatrixRows
                ),
                (
                    ref_base_image.TotalPixelMatrixColumns /
                    ref_image.TotalPixelMatrixColumns
                ),
            )
            pyramid_level = PyramidLevel(
                total_pixel_matrix_dimensions,
                pixel_spacing,
                downsampling_factors,
                False
            )
            key = (
                total_pixel_matrix_dimensions[0],
                total_pixel_matrix_dimensions[1],
            )
            level_lut[key] = pyramid_level
            level_keys.append(key)

        base_image = metadata[0]
        for i, image in enumerate(metadata):
            total_pixel_matrix_dimensions = (
                image.TotalPixelMatrixRows,
                image.TotalPixelMatrixColumns,
            )
            pixel_spacing = (
                float(
                    image
                    .SharedFunctionalGroupsSequence[0]
                    .PixelMeasuresSequence[0]
                    .PixelSpacing[1]
                ),
                float(
                    image
                    .SharedFunctionalGroupsSequence[0]
                    .PixelMeasuresSequence[0]
                    .PixelSpacing[0]
                ),
            )
            downsampling_factors = (
                (
                    base_image.TotalPixelMatrixRows /
                    image.TotalPixelMatrixRows
                ),
                (
                    base_image.TotalPixelMatrixColumns /
                    image.TotalPixelMatrixColumns
                ),
            )
            key = (
                total_pixel_matrix_dimensions[0],
                total_pixel_matrix_dimensions[1],
            )
            try:
                ref_pyramid_level = level_lut[key]
            except KeyError:
                raise ValueError(
                    f'Pyramid level #{i} does not match any level in the '
                    'reference pyramid. The reference pyramid has no level '
                    f'with total pixel matrix dimensions {key}.'
                )
            pyramid_level = PyramidLevel(
                total_pixel_matrix_dimensions,
                pixel_spacing,
                downsampling_factors,
                True
            )
            if not all([
                eq(
                    ref_pyramid_level.pixel_spacing,
                    pyramid_level.pixel_spacing
                ),
                eq(
                    ref_pyramid_level.downsampling_factors,
                    pyramid_level.downsampling_factors
                ),
            ]):
                raise ValueError(
                    f'Pyramid level #{i} does not match corresponding level '
                    'of reference pyramid. '
                    'Downsampling factors: '
                    f'{pyramid_level.downsampling_factors} versus '
                    f'{ref_pyramid_level.downsampling_factors}. '
                    'Pixel spacing: '
                    f'{pyramid_level.pixel_spacing} versus '
                    f'{ref_pyramid_level.pixel_spacing}.'
                )
            level_lut[key] = pyramid_level

        self._levels = tuple([level_lut[key] for key in level_keys])
        self._current_index = 0

    def __repr__(self) -> str:
        parts = []
        for i, level in enumerate(self):
            parts.append(f'=== Pyramid Level {i} ===')
            if level is not None:
                parts.append(
                    'Total Pixel Matrix Rows/Columns: '
                    f'{level.total_pixel_matrix_dimensions}'
                )
                parts.append(
                    f'Pixel Spacing: {level.pixel_spacing}'
                )
                parts.append(
                    f'Downsampling Factors: {level.downsampling_factors}'
                )
        return '\n'.join(parts)

    def __getitem__(self, key: int) -> PyramidLevel:
        try:
            return self._levels[key]
        except KeyError:
            raise IndexError(f'Pyramid does not have level {key}.')

    def __delitem__(self, key: int) -> None:
        raise AttributeError('Cannot delete pyramid level.')

    def __setitem__(self, key: int, value: PyramidLevel) -> None:
        raise AttributeError('Cannot set pyramid level.')

    def __iter__(self) -> Iterator[PyramidLevel]:
        return self

    def __next__(self) -> PyramidLevel:
        key = self._current_index
        if key < len(self._levels):
            self._current_index += 1
        else:
            self._current_index = 0
            raise StopIteration
        return self[key]

    def __len__(self) -> int:
        return len(self._levels)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, self.__class__):
            return False
        if len(self) != len(item):
            return False
        for i in range(len(self)):
            level = self[i]
            try:
                other_level = item[i]
            except IndexError:
                return False
            if level not in other_level:
                return False
        return True

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            level = self[i]
            try:
                other_level = other[i]
            except IndexError:
                return False
            if level != other_level:
                return False
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import re
from pathlib import Path

import setuptools

root_directory = Path(__file__).parent
readme_filepath = root_directory.joinpath('README.md')
long_description = readme_filepath.read_text()

version_filepath = root_directory.joinpath('src/dicomslide/version.py')
with io.open(version_filepath, 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)


setuptools.setup(
    name='dicomslide',
    version=version,
    description=(
        'Library for reading whole slide images and derived information '
        'in DICOM format.'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Markus D. Herrmann',
    maintainer='Markus D. Herrmann',
    url='https://github.com/herrmannlab/dicomslide',
    license='MIT',
    platforms=['Linux', 'MacOS', 'Windows'],
    classifiers=[
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Development Status :: 4 - Beta',
    ],
    include_package_data=True,
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.6',
    install_requires=[
        'dicomweb_client>=0.56.2',
        'highdicom>=0.16.0',
        'numpy>=1.19',
        'pydicom>=2.3.0',
        'scipy>=1.0',
    ],
)



import itertools
import logging
from collections import defaultdict, OrderedDict
from hashlib import sha256
from typing import (
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import highdicom as hd
import numpy as np
from dicomweb_client import DICOMClient
from pydicom import Dataset
from pydicom.sr.coding import Code
from pydicom.uid import (
    ParametricMapStorage,
    SegmentationStorage,
    VLWholeSlideMicroscopyImageStorage,
)

from dicomslide.enum import ChannelTypes
from dicomslide.image import TiledImage
from dicomslide.pyramid import get_image_size, Pyramid
from dicomslide.utils import (
    does_optical_path_item_match,
    does_segment_item_match,
    does_specimen_description_item_match,
    _encode_dataset,
    is_volume_image,
    is_label_image,
    is_overview_image,
)

logger = logging.getLogger(__name__)


class Slide:

    """A digital slide.

    A collection of DICOM image instances that share the same Frame of
    Reference UID and Container Identifier, i.e., that have been acquired as
    part of one image acquisition for the same physical glass slide (container)
    and can be visualized and analyzed in the same frame of reference
    (coordinate system).

    A slide consists of one or more image pyramids - one for each unique pair
    of channel and focal plane. The total pixel matrices of the different
    pyramid levels are stored in separate DICOM image instances. Individual
    channels or focal planes may be each stored in separate DICOM image
    instances or combined in a single DICOM image instance per pyramid level.
    Pyramids are expected to have the same number of levels and the same
    downsampling factors across channels and focal planes and the total pixel
    matrices at each level are expected to have the same dimensions (i.e., the
    same number of total pixel matrix columns and rows). However, the tiling of
    the total pixel matrices (i.e., the number of tile columns and rows) may
    differ across pyramid levels as well as across channels and focal planes at
    the same pyramid level.

    """

    def __init__(
        self,
        client: DICOMClient,
        image_metadata: Sequence[Dataset],
        max_frame_cache_size: int = 6,
        pyramid_tolerance: float = 0.1
    ):
        """

        Parameters
        ----------
        client: dicomweb_client.api.DICOMClient
            DICOMweb client
        image_metadata: Sequence[pydicom.Dataset]
            Metadata of DICOM VL Whole Slide Microscopy Image instances or of
            derived DICOM Segmentation or Parametric Map instances that belong
            to the slide
        max_frame_cache_size: int, optional
            Maximum number of frames that should be cached per image instance
            to avoid repeated retrieval requests
        pyramid_tolerance: float, optional
            Maximally tolerated distances between the centers of images at
            different pyramid levels in the slide coordinate system in
            millimeter unit

        """
        if not isinstance(image_metadata, Sequence):
            raise TypeError('Argument "image_metadata" must be a sequence.')
        if len(image_metadata) == 0:
            raise ValueError('Argument "image_metadata" cannot be empty.')

        logger.debug(f'construct Slide for n={len(image_metadata)} images')
        ref_image = image_metadata[0]
        volume_images_lut: Dict[
            Tuple[ChannelTypes, str, float], List[TiledImage]
        ] = defaultdict(list)
        label_images = []
        overview_images = []
        for i, metadata in enumerate(image_metadata):
            if not isinstance(metadata, Dataset):
                raise TypeError(
                    f'Item #{i} of argument "image_metadata" must have type '
                    'pydicom.Dataset.'
                )
            if metadata.FrameOfReferenceUID != ref_image.FrameOfReferenceUID:
                raise ValueError(
                    'All items of argument "image_metadata" must be DICOM '
                    'VL Whole Slide Microscpy Image instance with the same '
                    'Frame of Reference UID.'
                )
            if metadata.ContainerIdentifier != ref_image.ContainerIdentifier:
                raise ValueError(
                    'All items of argument "image_metadata" must be DICOM '
                    'VL Whole Slide Microscopy Image instance with the same '
                    'Container Identifier.'
                )
            image = TiledImage(
                client=client,
                image_metadata=metadata,
                max_frame_cache_size=max_frame_cache_size
            )
            if is_volume_image(metadata):
                iterator: Iterator[Tuple[int, int]] = itertools.product(
                    range(image.num_channels),
                    range(image.num_focal_planes),
                )
                for channel_index, focal_plane_index in iterator:
                    channel_type = image.channel_type
                    # FIXME: Segment Number may not be unique across
                    # Segmentation series/instances and would currently
                    # result in several images.
                    channel_identifier = image.get_channel_identifier(
                        channel_index
                    )
                    focal_plane_offset = image.get_focal_plane_offset(
                        focal_plane_index
                    )
                    key: Tuple[ChannelTypes, str, float] = (
                        channel_type,
                        channel_identifier,
                        focal_plane_offset,
                    )
                    volume_images_lut[key].append(image)
            if is_overview_image(metadata):
                overview_images.append(image)
            if is_label_image(metadata):
                label_images.append(image)

        if len(volume_images_lut) == 0:
            raise ValueError(
                'Slide must contain at least one VOLUME or THUMBNAIL image.'
            )
        self._label_images = tuple(label_images)
        self._overview_images = tuple(overview_images)
        self._volume_images: Dict[Tuple[int, int], Tuple[TiledImage, ...]] = {}

        logger.debug('assign images to channels')
        unique_channel_identifiers = defaultdict(set)
        unique_focal_plane_offsets = set()
        for key in volume_images_lut.keys():
            channel_type, channel_id, focal_plane_offset = key
            unique_channel_identifiers[channel_type].add(channel_id)
            unique_focal_plane_offsets.add(focal_plane_offset)

        self._number_of_channels = sum([
            len(ids) for ids in unique_channel_identifiers.values()
        ])
        self._channel_identifier_lut: Mapping[int, str] = OrderedDict()
        self._channel_type_lut: Mapping[int, ChannelTypes] = OrderedDict()
        self._channel_index_lut: Mapping[Tuple[ChannelTypes, str], int] = {}
        i = 0
        for channel_type, channel_ids in unique_channel_identifiers.items():
            for channel_id in sorted(channel_ids):
                self._channel_identifier_lut[i] = channel_id
                self._channel_type_lut[i] = channel_type
                self._channel_index_lut[(channel_type, channel_id)] = i
                i += 1

        self._number_of_focal_planes = len(unique_focal_plane_offsets)
        self._focal_plane_offset_lut: Mapping[int, float] = OrderedDict()
        self._focal_plane_index_lut: Mapping[float, int] = {}
        for i, focal_plane_offset in enumerate(
            sorted(unique_focal_plane_offsets)
        ):
            self._focal_plane_offset_lut[i] = focal_plane_offset
            self._focal_plane_index_lut[focal_plane_offset] = i

        ref_image_metadata: List[TiledImage] = []
        encoded_image_metadata = []
        for channel_index in self._channel_identifier_lut.keys():
            channel_id = self._channel_identifier_lut[channel_index]
            channel_type = self._channel_type_lut[channel_index]
            for focal_plane_index in self._focal_plane_offset_lut.keys():
                focal_plane_offset = self._focal_plane_offset_lut[
                    focal_plane_index
                ]
                volume_images: List[TiledImage] = sorted(
                    volume_images_lut[
                        (channel_type, channel_id, focal_plane_offset)
                    ],
                    key=lambda image: get_image_size(image.metadata),
                    reverse=True
                )
                volume_image_key: Tuple[int, int] = (
                    channel_index,
                    focal_plane_index,
                )
                self._volume_images[volume_image_key] = tuple(volume_images)
                if (
                    channel_type == ChannelTypes.OPTICAL_PATH and
                    len(ref_image_metadata) == 0
                ):
                    ref_image_metadata.extend(volume_images)
                encoded_image_metadata.extend([
                    _encode_dataset(image.metadata)
                    for images in volume_images
                ])
        encoded_image_metadata.extend([
            _encode_dataset(image.metadata)
            for image in overview_images + label_images
        ])

        logger.debug('build pyramids for each channel')
        pyramids: Dict[Tuple[int, int], Pyramid] = {}
        for (
            channel_index,
            focal_plane_index,
        ), image_collection in self._volume_images.items():
            if len(ref_image_metadata) > 0:
                ref_metadata = [
                    image.metadata
                    for image in ref_image_metadata
                ]
            else:
                ref_metadata = None
            try:
                pyramids[(channel_index, focal_plane_index)] = Pyramid(
                    metadata=[
                        image.metadata
                        for image in image_collection
                    ],
                    tolerance=pyramid_tolerance,
                    ref_metadata=ref_metadata
                )
            except ValueError as error:
                raise ValueError(
                    f'VOLUME and THUMBNAIL images for channel {channel_index} '
                    f'and focal plane {focal_plane_index} do not represent '
                    'a valid image pyramid. '
                    f'{error}'
                )

        # For now, pyramids must be identical across channels and focal planes.
        # This requirement could potentially be relaxed in the future.
        ref_pyramid = pyramids[(0, 0)]
        for pyramid in pyramids.values():
            if pyramid not in ref_pyramid:
                raise ValueError(
                    'Pyramids for different channels and focal planes must '
                    'have the same structure, i.e., the same number of levels '
                    'as well as the same dimensions and downsampling factors '
                    'per level.'
                )
        self._pyramid = ref_pyramid

        # The hash is computed using the image metadata rather than the pixel
        # data to avoid having to retrieve the potentially large pixel data.
        self._quickhash = sha256(b''.join(encoded_image_metadata)).hexdigest()

    def __repr__(self) -> str:
        ref_images = self._volume_images[(0, 0)]
        metadata = ref_images[0].metadata
        return f'<Slide {metadata.ContainerIdentifier} {self._quickhash}>'

    def __hash__(self) -> int:
        return hash(self._quickhash)

    def find_segments(
        self,
        number: Optional[int] = None,
        label: Optional[str] = None,
        property_category: Optional[Union[hd.sr.CodedConcept, Code]] = None,
        property_type: Optional[Union[hd.sr.CodedConcept, Code]] = None
    ) -> Tuple[int, ...]:
        """Find segments.

        Parameters
        ----------
        number: Union[int, None], optional
            Segment number
        label: Union[str, None], optional,
            Segment label
        property_category: Union[hd.sr.CodedConcept, Code, None], optional
            Category of segmented property
        property_type: Union[hd.sr.CodedConcept, Code, None], optional
            Type of segmented property

        Returns
        -------
        Tuple[int, ...]
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.

        """
        matching_channel_indices = set()
        for (channel_index, _), images in self._volume_images.items():
            ref_image = images[0]
            if ref_image.channel_type != ChannelTypes.SEGMENT:
                continue
            channel_identifier = self.get_channel_identifier(channel_index)
            if number is not None:
                if channel_identifier == str(number):
                    return (channel_index, )

            matching_segment_items = [
                item
                for item in ref_image.metadata.SegmentSequence
                if channel_identifier == str(item.SegmentNumber)
            ]
            segment_item = matching_segment_items[0]
            if does_segment_item_match(
                segment_item,
                number,
                label,
                property_category,
                property_type,
            ):
                matching_channel_indices.add(channel_index)

        return tuple(matching_channel_indices)

    def find_optical_paths(
        self,
        identifier: Optional[str] = None,
        description: Optional[str] = None,
        illumination_wavelength: Optional[float] = None,
        specimen_stain: Optional[Union[hd.sr.CodedConcept, Code]] = None
    ) -> Tuple[int, ...]:
        """Find optical paths.

        Parameters
        ----------
        identifier: Union[str, None], optional
            Optical path identifier
        description: Union[str, None], optional,
            Optical path description
        illumination_wavelength: Union[float, None], optional,
            Optical path illumination wavelength
        specimen_stain: Union[hd.sr.CodedConcept, Code, None], optional
            Substance used for specimen staining

        Returns
        -------
        Tuple[int, ...]
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.

        """
        matching_channel_indices = set()
        for (channel_index, _), images in self._volume_images.items():
            ref_image = images[0]
            if ref_image.channel_type != ChannelTypes.OPTICAL_PATH:
                continue
            channel_identifier = self.get_channel_identifier(channel_index)
            if identifier is not None:
                if channel_identifier == identifier:
                    return (channel_index, )

            matching_optical_path_items = [
                item
                for item in ref_image.metadata.OpticalPathSequence
                if item.OpticalPathIdentifier == channel_identifier
            ]
            optical_path_item = matching_optical_path_items[0]
            specimen_description_item = (
                ref_image
                .metadata
                .SpecimenDescriptionSequence[0]
            )
            if all([
                does_optical_path_item_match(
                    optical_path_item,
                    identifier,
                    description,
                    illumination_wavelength
                ),
                does_specimen_description_item_match(
                    specimen_description_item,
                    specimen_stain
                )
            ]):
                matching_channel_indices.add(channel_index)

        return tuple(matching_channel_indices)

    def get_volume_images(
        self,
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> Tuple[TiledImage, ...]:
        """Get VOLUME or THUMBNAIL images for an channel and focal plane.

        Parameters
        ----------
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        Tuple[dicomslide.TiledImage, ...]
            Images sorted by size in descending order

        """
        key = (channel_index, focal_plane_index)
        try:
            return tuple(self._volume_images[key])
        except KeyError:
            raise IndexError(
                f'No VOLUME images found for channel {channel_index} '
                f'and focal plane {focal_plane_index}.'
            )

    @property
    def label_images(self) -> Tuple[TiledImage, ...]:
        """Tuple[dicomslide.TiledImage, ...]: LABEL images of the slide"""
        return self._label_images

    @property
    def overview_images(self) -> Tuple[TiledImage, ...]:
        """Tuple[dicomslide.TiledImage, ...]: OVERVIEW images of the slide"""
        return self._overview_images

    @property
    def num_channels(self) -> int:
        """int: Number of channels"""
        return self._number_of_channels

    def get_channel_type(self, channel_index: int) -> ChannelTypes:
        """Get type of a channel.

        Parameters
        ----------
        channel_index: int
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.

        Returns
        -------
        dicomslide.ChannelTypes
            Channel type

        Raises
        ------
        ValueError
            When no channel is found for `channel_index`

        """
        try:
            return self._channel_type_lut[channel_index]
        except IndexError:
            raise ValueError(
                'No VOLUME or THUMNAIL image found for channel index '
                f'{channel_index}.'
            )

    def get_channel_identifier(self, channel_index: int) -> str:
        """Get identifier of a channel.

        Parameters
        ----------
        channel_index: int
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.

        Returns
        -------
        str
            Channel identifier

        Raises
        ------
        ValueError
            When no channel is found for `channel_index`

        """
        try:
            return self._channel_identifier_lut[channel_index]
        except IndexError:
            raise ValueError(
                'No VOLUME or THUMNAIL image found for channel index '
                f'{channel_index}.'
            )

    def get_channel_index(
        self,
        channel_identifier: str,
        channel_type: Union[ChannelTypes, str]
    ) -> int:
        """Get index of a channel.

        Parameters
        ----------
        channel_identifier: str
            Channel identifier
        channel_type: Union[str, dicomslide.ChannelTypes]
            Channel type

        Returns
        -------
        int
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute, which is
            dependend on the type of channel.

        Raises
        ------
        ValueError
            When no channel is found for `channel_identifier` and `channel_type`

        """
        channel_type = ChannelTypes(channel_type)
        try:
            return self._channel_index_lut[(channel_type, channel_identifier)]
        except IndexError:
            raise ValueError(
                'No VOLUME or THUMNAIL image found for channel identifier '
                f'{channel_identifier} and type {channel_type.value}.'
            )

    @property
    def frame_of_reference_uid(self) -> str:
        """str: Unique identifier of the frame of reference"""
        volume_images = self.get_volume_images(
            channel_index=0,
            focal_plane_index=0
        )
        return volume_images[0].frame_of_reference_uid

    @property
    def physical_offset(self) -> Tuple[float, float]:
        """Tuple[float, float]: Minimum offset of the total pixel matrices
        from the origin of the frame of reference along the X and Y axes of the
        slide coordinate system in millimeter

        """
        offsets = np.array([
            volume_images[0].physical_offset
            for volume_images in self._volume_images.values()
        ])
        min_offsets = np.min(offsets, axis=0)
        return (min_offsets[0], min_offsets[1])

    @property
    def size(self) -> Tuple[int, int]:
        """Tuple[int, int]: Maximum size of the total pixel matrices along
        the rows and columns axes of the total pixel matrix

        """
        sizes = np.array([
            volume_images[0].size
            for volume_images in self._volume_images.values()
        ])
        max_sizes = np.max(sizes, axis=0)
        return (max_sizes[0], max_sizes[1])

    @property
    def physical_size(self) -> Tuple[float, float]:
        """Tuple[float, float]: Maximum size of the total pixel matrices along
        the X and Y axes of the slide coordinate system in millimeter

        """
        x_offset, y_offset = self.physical_offset
        sizes = []
        for volume_images in self._volume_images.values():
            image = volume_images[0]
            x_endpoint, y_endpoint = image.get_slide_offset(image.size)
            sizes.append(
                (
                    abs(x_endpoint - x_offset),
                    abs(y_endpoint - y_offset),
                )
            )
        max_sizes = np.max(np.array(sizes), axis=0)
        return (max_sizes[0], max_sizes[1])

    @property
    def num_focal_planes(self) -> int:
        """int: Number of focal planes"""
        return self._number_of_focal_planes

    def get_focal_plane_offset(self, focal_plane_index: int) -> float:
        """Get z offset of focal plane in slide coordinate system.

        Parameters
        ----------
        focal_plane_index: int
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        float
            Offset of the focal plane from the from the slide surface along the
            Z axis of the slide coordinate system in micrometers

        Raises
        ------
        ValueError
            When no focal plane is found for `focal_plane_index`

        """
        try:
            return self._focal_plane_offset_lut[focal_plane_index]
        except IndexError:
            raise ValueError(
                'No VOLUME or THUMNAIL image found for focal plane index '
                f'{focal_plane_index}.'
            )

    def get_focal_plane_index(self, focal_plane_offset: float) -> int:
        """Get index of a focal plane.

        Parameters
        ----------
        focal_plane_offset: float
            Offset of the focal plane from the from the slide surface along the
            Z axis of the slide coordinate system in micrometers

        Returns
        -------
        int
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Raises
        ------
        ValueError
            When no focal plane is found for `focal_plane_offset`

        """
        try:
            return self._focal_plane_index_lut[focal_plane_offset]
        except IndexError:
            raise ValueError(
                'No VOLUME or THUMNAIL image found for focal plane offset '
                f'{focal_plane_offset}.'
            )

    @property
    def num_levels(self) -> int:
        """int: Number of pyramid levels

        Note
        ----
        Levels are sorted by size in descending order from the base level
        (highest image resolution, smallest pixel spacing) to the top level
        (lowest image resolution, largest pixel spacing).

        """
        return len(self._pyramid)

    @property
    def total_pixel_matrix_dimensions(self) -> Tuple[Tuple[int, int], ...]:
        """Tuple[Tuple[int, int], ...]: Number of columns and rows in the total
        pixel matrix for images at each pyramid level

        """
        return tuple([
            level.total_pixel_matrix_dimensions
            for level in self._pyramid
        ])

    @property
    def pixel_spacings(self) -> Tuple[Tuple[float, float], ...]:
        """Tuple[Tuple[float, float], ...]: Distance between neighboring pixels
        along the row (left to right) and column (top to bottom) directions

        """
        return tuple([level.pixel_spacing for level in self._pyramid])

    @property
    def downsampling_factors(self) -> Tuple[float, ...]:
        """Tuple[float]: Downsampling factors of images at each pyramid level
        relative to the base level

        """
        return tuple([
            float(np.mean(level.downsampling_factors))
            for level in self._pyramid
        ])

    def map_pixel_indices_to_slide_coordinates(
        self,
        pixel_indices: np.ndarray,
        level: int,
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> np.ndarray:
        """Map pixel indices to slide coordinates.

        Parameters
        ----------
        pixel_indices: numpy.ndarray
            Zero-based (row, column) indices into the total pixel matrix of the
            image
        level: int
            Zero-based index into pyramid levels
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        numpy.ndarray
            Zero-based (x, y, z) coordinates in the slide coordinate system in
            millimeter

        """
        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')
        return image.map_pixel_indices_to_slide_coordinates(pixel_indices)

    def map_slide_coordinates_to_pixel_indices(
        self,
        slide_coordinates: np.ndarray,
        level: int,
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> np.ndarray:
        """Map slide coordinates to pixel indices.

        Parameters
        ----------
        slide_coordinates: numpy.ndarray
            Zero-based (x, y, z) coordinates in the slide coordinate system in
            millimeter
        level: int
            Zero-based index into pyramid levels
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        numpy.ndarray
            Zero-based (row, column) indices into the total pixel matrix of the
            image

        """
        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')
        return image.map_slide_coordinates_to_pixel_indices(slide_coordinates)

    def get_slide_offset(
        self,
        pixel_indices: Tuple[int, int],
        level: int,
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> Tuple[float, float]:
        """Get slide coordinates for a given total pixel matrix position.

        Parameters
        ----------
        pixel_indices: Tuple[int, int]
            Zero-based (row, column) offset in the total pixel matrix
        level: int
            Zero-based index into pyramid levels
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        Tuple[float, float]
            Zero-based (x, y) position on the slide in the slide coordinate
            system in millimeter

        """
        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')
        return image.get_slide_offset(pixel_indices)

    def get_pixel_indices(
        self,
        offset: Tuple[float, float],
        level: int,
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> Tuple[int, int]:
        """Get indices into total pixel matrix for a given slide position.

        Parameters
        ----------
        offset: Tuple[float, float]
            Zero-based (x, y) offset in the slide coordinate system in
            millimeter
        level: int
            Zero-based index into pyramid levels
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        Tuple[int, int]
            Zero-based (row, column) position in the total pixel matrix of the
            image

        Note
        ----
        Pixel position may be negativ or extend beyond the size of the total
        pixel matrix if slide position at `offset` does fall into a region on
        the slide that was not imaged.

        """
        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')
        return image.get_pixel_indices(offset)

    def get_image_region(
        self,
        offset: Tuple[int, int],
        level: int,
        size: Tuple[int, int],
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> np.ndarray:
        """Get image region.

        Parameters
        ----------
        offset: Tuple[int, int]
            Zero-based (row, column) indices in the range [0, Rows) and
            [0, Columns), respectively, that specify the offset of the image
            region in the total pixel matrix of the image at the highest
            resolution level. The ``(0, 0)`` coordinate is located at the
            center of the topleft hand pixel in the total pixel matrix.
        level: int
            Zero-based index into pyramid levels
        size: Tuple[int, int]
            Rows and columns of the requested image region
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        numpy.ndarray
            Three-dimensional pixel array of shape
            (Rows, Columns, Samples per Pixel) for the requested image region

        """
        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')

        row_index, col_index = offset
        col_factor, row_factor = self._pyramid[level].downsampling_factors
        col_start = int(np.floor(col_index / col_factor))
        row_start = int(np.floor(row_index / row_factor))
        image_offset = (row_start, col_start)

        # Each image may have one or more channels or focal planes and the
        # image-level indices may differ from the slide-level indices.
        image_channel_index = image.get_channel_index(
            self.get_channel_identifier(channel_index)
        )
        image_focal_plane_index = image.get_focal_plane_index(
            self.get_focal_plane_offset(focal_plane_index)
        )
        return image.get_image_region(
            offset=image_offset,
            size=size,
            channel_index=image_channel_index,
            focal_plane_index=image_focal_plane_index
        )

    def get_slide_region(
        self,
        offset: Tuple[float, float],
        level: int,
        size: Tuple[float, float],
        channel_index: int = 0,
        focal_plane_index: int = 0
    ) -> np.ndarray:
        """Get slide region.

        Parameters
        ----------
        offset: Tuple[float, float]
            Zero-based (x, y) offset in the slide coordinate system in
            millimeter resolution. The ``(0.0, 0.0)`` coordinate is located at
            the origin of the slide (usually the slide corner).
        level: int
            Zero-based index into pyramid levels
        size: Tuple[float, float]
            Width and height of the requested slide region in millimeter unit
            along the X and Y axis of the slide coordinate system, respectively.
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        focal_plane_index: int, optional
            Zero-based index into focal planes along depth direction from the
            glass slide towards the coverslip in the slide coordinate system
            specified by the Z Offset in Slide Coordinate System attribute of
            VOLUME or THUMBNAIL images.

        Returns
        -------
        numpy.ndarray
            Three-dimensional pixel array of shape
            (Rows, Columns, Samples per Pixel) for the requested slide region

        Note
        ----
        The slide coordinate system is defined for the upright standing slide
        such that the X axis corresponds to the short side of the slide and the
        Y axis corresponds to the long side of the slide.
        The rows of the returned pixel array are thus parallel to the X axis of
        the slide coordinate system and the columns parallel to the Y axis of
        the slide coordinate system.

        """
        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')

        # Each image may have one or more channels or focal planes and the
        # image-level indices may differ from the slide-level indices.
        image_channel_index = image.get_channel_index(
            self.get_channel_identifier(channel_index)
        )
        image_focal_plane_index = image.get_focal_plane_index(
            self.get_focal_plane_offset(focal_plane_index)
        )
        return image.get_slide_region(
            offset=offset,
            size=size,
            channel_index=image_channel_index,
            focal_plane_index=image_focal_plane_index
        )

    def get_slide_region_for_annotation(
        self,
        annotation: hd.sr.Scoord3DContentItem,
        level: int,
        channel_index: int = 0,
        padding: Union[
            float,
            Tuple[float, float],
            Tuple[float, float, float, float]
        ] = 0.0
    ) -> np.ndarray:
        """Get slide region defined by a graphic annotation.

        Parameters
        ----------
        annotation: highdicom.sr.Scoord3DContentItem
            Graphic annotation that defines the region of interest (ROI) in the
            slide coordinate system
        level: int
            Zero-based index into pyramid levels
        channel_index: int, optional
            Zero-based index into channels along the direction defined by
            successive items of the appropriate DICOM attribute of VOLUME
            or THUMBNAIL images.
        padding: Union[int, Tuple[int, int], Tuple[int, int, int, int]], optional
            Padding on each border of the region defined by `annotation`.  If a
            single integer is provided, the value is used to pad all four
            borders with the same number of pixels. If a sequence of length 2
            is provided, the two values are used to pad the left/right (along
            the X axis) and top/bottom (along the Y axis) border, respectively.
            If a sequence of length 4 is provided, the four values are used to
            pad the left (- X axis), top (+ Y axis), right (+ X axis), and
            bottom (- Y axis) borders respectively.

        Returns
        -------
        numpy.ndarray
            Three-dimensional pixel array of shape
            (Rows, Columns, Samples per Pixel) for the requested slide region

        Note
        ----
        The slide coordinate system is defined for the upright standing slide
        such that the X axis corresponds to the short side of the slide and the
        Y axis corresponds to the long side of the slide.
        The rows of the returned pixel array are thus parallel to the X axis of
        the slide coordinate system and the columns parallel to the Y axis of
        the slide coordinate system.

        """  # noqa: E501
        full_padding: Tuple[float, float, float, float]
        if isinstance(padding, (int, float)):
            full_padding = (
                float(padding),
                float(padding),
                float(padding),
                float(padding),
            )
        elif isinstance(padding, tuple):
            if len(padding) == 2:
                full_padding = (
                    float(padding[0]),
                    float(padding[1]),
                    float(padding[0]),
                    float(padding[1]),
                )
            elif len(padding) == 4:
                full_padding = (
                    float(padding[0]),
                    float(padding[1]),
                    float(padding[2]),  # type: ignore
                    float(padding[3]),  # type: ignore
                )
            else:
                raise ValueError(
                    'If argument "padding" is a tuple, its length must be '
                    'either 2 or 4.'
                )
        else:
            raise TypeError(
                'Argument "padding" must be either an integer or a tuple.'
            )

        slide_coordinates: Tuple[float, float]
        size: Tuple[float, float]
        focal_plane_offset: float
        graphic_data = annotation.value
        if annotation.graphic_type == hd.sr.GraphicTypeValues3D.POINT:
            offset = (graphic_data[0, 0], graphic_data[0, 1])
            focal_plane_offset = graphic_data[0, 2]
            size = (0.0, 0.0)
        elif annotation.graphic_type in (
            hd.sr.GraphicTypeValues3D.MULTIPOINT,
            hd.sr.GraphicTypeValues3D.POLYGON,
            hd.sr.GraphicTypeValues3D.POLYLINE,
        ):
            min_point = np.min(graphic_data, axis=0)
            max_point = np.max(graphic_data, axis=0)
            offset = (min_point[0], min_point[1])
            focal_plane_offset = min_point[2]
            size = (max_point[0] - min_point[0], max_point[1] - min_point[1])
        elif annotation.graphic_type == hd.sr.GraphicTypeValues3D.ELLIPSE:
            min_point = [
                np.min([graphic_data[:, 0]]),
                np.min([graphic_data[:, 1]]),
            ]
            max_point = [
                np.max([graphic_data[:, 0]]),
                np.max([graphic_data[:, 1]]),
            ]
            offset = (min_point[0], min_point[1])
            # Points are co-planar per definition
            focal_plane_offset = graphic_data[0, 2]
            size = (max_point[0] - min_point[0], max_point[1] - min_point[1])
        elif annotation.graphic_type == hd.sr.GraphicTypeValues3D.ELLIPSOID:
            raise ValueError(
                'SCOORD3D graphic type "ELLIPSOID" is not supported, '
                'because annotations must be planar and the plane must be '
                'parallel to the slide surface.'
            )
        else:
            raise ValueError(
                'Unknown SCOORD3D graphic type "{annotation.graphic_type}".'
            )

        # TODO: Instead of expecting an exact match, one could select the
        # focal plane closest to the annotated region of interest (ROI).
        # The coordinates of the ROI are expected to be co-planar and parallel
        # to the slide surface but they may not be located on the same plane
        # as any of the available focal planes and may instead need to be
        # projected onto the nearest focal plane.
        focal_plane_index = self.get_focal_plane_index(focal_plane_offset)

        volume_images = self.get_volume_images(
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )
        try:
            image = volume_images[level]
        except IndexError:
            raise IndexError(f'Slide does not have level {level}.')
        frame_of_reference_uid = image.metadata.FrameOfReferenceUID
        ref_frame_of_reference_uid = annotation.ReferencedFrameOfReferenceUID
        if ref_frame_of_reference_uid != frame_of_reference_uid:
            raise ValueError(
                'Annotation must be defined in same frame of reference as the '
                f'source images: "{frame_of_reference_uid}".'
            )

        offset = (
            offset[0] - full_padding[0],
            offset[1] + full_padding[1],  # offset increases along Y axis!
        )
        size = (
            size[0] + full_padding[0] + full_padding[2],
            size[1] + full_padding[1] + full_padding[3],
        )

        return self.get_slide_region(
            offset=offset,
            level=level,
            size=size,
            channel_index=channel_index,
            focal_plane_index=focal_plane_index
        )


def find_slides(
    client: DICOMClient,
    study_instance_uid: Optional[str] = None,
    patient_id: Optional[str] = None,
    study_id: Optional[str] = None,
    container_id: Optional[str] = None,
    max_frame_cache_size: int = 6,
    pyramid_tolerance: float = 0.1,
    fail_on_error: bool = True,
    include_derived: bool = True,
    specimen_stains: Optional[Sequence[Union[Code, hd.sr.CodedConcept]]] = None,
    optical_path_ids: Optional[Sequence[str]] = None
) -> List[Slide]:
    """Find slides.

    Parameters
    ----------
    client: dicomweb_client.api.DICOMClient
        DICOMweb client
    study_instance_uid: Union[str, None], optional
        DICOM Study Instance UID
    patient_id: Union[str, None], optional
        Patient identifier
    study_id: Union[str, None], optional
        Study identifier
    container_id: Union[str, None], optional
        Specimen container (slide) identifier
    max_frame_cache_size: int, optional
        Maximum number of frames that should be cached per image instance to
        avoid repeated retrieval requests
    pyramid_tolerance: float, optional
        Maximally tolerated distances between the centers of images at
        different pyramid levels in the slide coordinate system in
        millimeter unit
    fail_on_error: bool, optional
        Wether the function should raise an exception in case an error occurs.
        If ``False``, slides will be skipped.
    include_derived: bool, optional
        Whether derived images (DICOM Segmentation or DICOM Parametric Map
        instances) should be considered and included into slides
    specimen_stains: Union[Sequence[Union[pydicom.sr.Code, highdicom.sr.CodedConcept]], None]
        Specimen stains for which corresponding slide microscopy images should
        be included in slides. Any image that does not contain one of the
        stains in the speciment description will be omitted.
    optical_path_ids: Union[Sequence[str], None]
        Identifiers of optical paths for which corresponding slide microscopy
        images should be included in slides. Any image that does not contain
        any of the specified optical paths will be omitted.

    Returns
    -------
    List[dicomslide.Slide]
        Digital slides

    """  # noqa: E501
    def bulk_data_uri_handler(
        tag: str,
        vr: str,
        uri: str
    ) -> Union[bytes, None]:
        # Only retrieve ICC Profile
        if tag != '00282000':
            return None
        return client.retrieve_bulkdata(uri)[0]

    # Find VL Whole Slide Microscopy Image instances
    logger.debug('search for slide microscopy image instances')
    search_filters: Dict[str, str] = {}
    if study_instance_uid is not None:
        search_filters['StudyInstanceUID'] = study_instance_uid
    if patient_id is not None:
        search_filters['PatientID'] = patient_id
    if study_id is not None:
        search_filters['StudyID'] = study_id

    # Search for instances per study to speed up subsequent queries.
    if len(search_filters) > 0:
        study_search_results = [
            Dataset.from_json(ds)
            for ds in client.search_for_studies(
                search_filters=search_filters,
                get_remaining=True
            )
        ]
    else:
        study_search_results = [
            Dataset.from_json(ds)
            for ds in client.search_for_studies(get_remaining=True)
        ]
    if len(study_search_results) == 0:
        return []

    lut = defaultdict(list)
    for study in study_search_results:
        # These checks should not be necessary, because we are using these
        # attributes to search for studies in the first place. However,
        # some archives get this wrong. Safety first!
        if study_instance_uid is not None:
            if study.StudyInstanceUID != study_instance_uid:
                logger.debug(
                    f'skip study "{study.SOPInstanceUID}" because it '
                    'does not match the Study Instance UID '
                    f'"{study_instance_uid}"'
                )
                continue
        if patient_id is not None:
            if study.PatientID != patient_id:
                logger.debug(
                    f'skip study "{study.SOPInstanceUID}" because it '
                    f'does not match the Patient ID "{patient_id}"'
                )
                continue
        if study_id is not None:
            if study.StudyID != study_id:
                logger.debug(
                    f'skip image "{study.SOPInstanceUID}" because it '
                    f'does not match the Study ID "{study_id}"'
                )
                continue

        current_study_instance_uid = study.StudyInstanceUID
        instance_search_results = []
        # We could search by SOPClassUID directly, but some archives don't
        # support this attribute as a query parameter (although they should
        # according to the standard).
        sm_image_instance_search_results = [
            Dataset.from_json(ds)
            for ds in client.search_for_instances(
                study_instance_uid=current_study_instance_uid,
                search_filters={'Modality': 'SM'}
            )
        ]
        sm_image_instance_search_results = [
            instance
            for instance in sm_image_instance_search_results
            if instance.SOPClassUID == VLWholeSlideMicroscopyImageStorage
        ]
        logger.debug(
            f'found n={len(sm_image_instance_search_results)} '
            'slide microscopy image instances '
            f'for study "{study.StudyInstanceUID}"'
        )
        instance_search_results.extend(sm_image_instance_search_results)

        if include_derived:
            seg_image_instance_search_results = [
                Dataset.from_json(ds)
                for ds in client.search_for_instances(
                    study_instance_uid=current_study_instance_uid,
                    search_filters={'Modality': 'SEG'}
                )
            ]
            seg_image_instance_search_results = [
                instance
                for instance in seg_image_instance_search_results
                if instance.SOPClassUID == SegmentationStorage
            ]
            logger.debug(
                f'found n={len(seg_image_instance_search_results)} '
                'segmentation image instances '
                f'for study "{study.StudyInstanceUID}"'
            )
            instance_search_results.extend(seg_image_instance_search_results)

            pm_image_instance_search_results = [
                Dataset.from_json(ds)
                for ds in client.search_for_instances(
                    study_instance_uid=current_study_instance_uid,
                    search_filters={'Modality': 'OT'}
                )
            ]
            pm_image_instance_search_results = [
                instance
                for instance in sm_image_instance_search_results
                if instance.SOPClassUID == ParametricMapStorage
            ]
            logger.debug(
                f'found n={len(pm_image_instance_search_results)} '
                'parametric map image instances '
                f'for study "{study.StudyInstanceUID}"'
            )
            instance_search_results.extend(pm_image_instance_search_results)
            logger.debug(
                f'found n={len(instance_search_results)} image instances '
                f'for study "{study.StudyInstanceUID}"'
            )

        logger.debug('filter image instances based on metadata')
        for instance in instance_search_results:
            logger.debug(
                'retrieve metadata for '
                f'image instance "{instance.SOPInstanceUID}"'
            )
            metadata = Dataset.from_json(
                client.retrieve_instance_metadata(
                    study_instance_uid=current_study_instance_uid,
                    series_instance_uid=instance.SeriesInstanceUID,
                    sop_instance_uid=instance.SOPInstanceUID,
                ),
                bulk_data_uri_handler=bulk_data_uri_handler
            )
            # Once more, these checks should not be necessary, because we are
            # using these attributes to search for studies in the first place.
            # However, some archives get this wrong. Safety first!
            if study_instance_uid is not None:
                if metadata.StudyInstanceUID != study_instance_uid:
                    logger.debug(
                        f'skip image "{metadata.SOPInstanceUID}" because it '
                        'does not match the Study Instance UID '
                        f'"{study_instance_uid}"'
                    )
                    continue
            if patient_id is not None:
                if metadata.PatientID != patient_id:
                    logger.debug(
                        f'skip image "{metadata.SOPInstanceUID}" because it '
                        f'does not match the Patient ID "{patient_id}"'
                    )
                    continue
            if study_id is not None:
                if metadata.StudyID != study_id:
                    logger.debug(
                        f'skip image "{metadata.SOPInstanceUID}" because it '
                        f'does not match the Study ID "{study_id}"'
                    )
                    continue

            try:
                current_container_id = metadata.ContainerIdentifier
            except AttributeError:
                logger.debug(
                    f'skip image "{metadata.SOPInstanceUID}" because it '
                    'does not have a Container Identifier attribute'
                )
                continue

            if container_id is not None:
                if current_container_id != container_id:
                    logger.debug(
                        f'skip image "{metadata.SOPInstanceUID}" because it '
                        f'does not match Container Identifier "{container_id}"'
                    )
                    continue

            try:
                current_frame_of_reference_uid = metadata.FrameOfReferenceUID
            except AttributeError:
                logger.warning(
                    f'skip image "{metadata.SOPInstanceUID}" because it '
                    'does not have a Frame of Reference UID attribute'
                )
                continue

            if optical_path_ids is not None:
                if instance.SOPClassUID == VLWholeSlideMicroscopyImageStorage:
                    does_match = False
                    for identifier in optical_path_ids:
                        for item in metadata.OpticalPathSequence:
                            does_match |= does_optical_path_item_match(
                                item=item,
                                identifier=identifier
                            )
                    if not does_match:
                        logger.debug(
                            f'skip image "{metadata.SOPInstanceUID}" because '
                            'its optical path does not match any of '
                            'the specified optical path identitiers'
                        )
                        continue

            if specimen_stains is not None:
                if instance.SOPClassUID == VLWholeSlideMicroscopyImageStorage:
                    does_match = False
                    for stain in specimen_stains:
                        for item in metadata.SpecimenDescriptionSequence:
                            does_match |= does_specimen_description_item_match(
                                item=item,
                                specimen_stain=stain
                            )
                    if not does_match:
                        logger.debug(
                            f'skip image "{metadata.SOPInstanceUID}" because '
                            'its specimen description does not match any of '
                            'the specified specimen stains'
                        )
                        continue

            logger.debug(
                f'assign image "{metadata.SOPInstanceUID}" to slide with '
                f'Study Instance UID "{current_study_instance_uid}", '
                f'Container Identifier "{current_container_id}", and '
                f'Frame of Reference UID "{current_frame_of_reference_uid}"'
            )
            key = (
                current_study_instance_uid,
                current_container_id,
                current_frame_of_reference_uid,
            )
            lut[key].append(metadata)

    logger.debug(f'create n={len(lut)} slide objects')
    found_slides = []
    for _, image_metadata in lut.items():
        ref_image = image_metadata[0]
        logger.debug(
            f'create slide for images of study "{ref_image.StudyInstanceUID}" '
            f'for container "{ref_image.ContainerIdentifier}" '
            f'in frame of reference "{ref_image.FrameOfReferenceUID}"'
        )
        try:
            slide = Slide(
                client=client,
                image_metadata=image_metadata,
                max_frame_cache_size=max_frame_cache_size,
                pyramid_tolerance=pyramid_tolerance
            )
            found_slides.append(slide)
        except Exception as error:
            if fail_on_error:
                raise
            else:
                ref_image = image_metadata[0]
                logger.warning(
                    'failed to create slide for images of study '
                    f'"{ref_image.StudyInstanceUID}" for container '
                    f'"{ref_image.ContainerIdentifier}" in frame of reference '
                    f'"{ref_image.FrameOfReferenceUID}" due to the following '
                    f'error: {error}'
                )

    return found_slides
